# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.projection import Projection
from moleculekit.periodictable import periodictable
import numpy as np
import logging

logger = logging.getLogger(__name__)


class MetricGyration(Projection):
    """Creates a MetricGyration object that calculates the radius of gyration (ROG) of a molecule.

    Parameters
    ----------
    atomsel : str
        Atom selection string for the atoms whose ROG we want to calculate.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__

    Returns
    -------
    metr : MetricGyration object
    """

    def __init__(
        self,
        atomsel,
        refmol=None,
        trajalnsel=None,
        refalnsel=None,
        centersel="protein",
        pbc=True,
    ):
        super().__init__()

        if atomsel is None:
            raise ValueError("Atom selection cannot be None")

        self._atomsel = atomsel
        self._trajalnsel = trajalnsel
        self._centersel = centersel
        self._pbc = pbc
        self._refmol = refmol

        if self._refmol is not None:
            if self._trajalnsel is None:
                self._trajalnsel = "protein and name CA"
            self._refalnsel = refalnsel if refalnsel is not None else self._trajalnsel
            self._refmol = refmol.copy()
            self._cache["refalnsel"] = self._refmol.atomselect(self._refalnsel)

    def _calculateMolProp(self, mol, props="all"):
        props = ("atomsel", "masses") if props == "all" else props
        res = {}
        if "atomsel" in props:
            res["atomsel"] = mol.atomselect(self._atomsel)
            if np.sum(res["atomsel"]) == 0:
                raise RuntimeError("Atom selection resulted in 0 atoms.")

        if "masses" in props:
            sel = mol.atomselect(self._atomsel)
            res["masses"] = mol.masses[sel]
            if np.any(res["masses"] == 0):
                logger.warning(
                    "The Molecule has atoms with 0 mass. Guessing the masses from the elements."
                )
                res["masses"] = np.array(
                    [periodictable[el].mass for el in mol.element[sel]]
                )
                if np.sum(res["masses"]) == 0:
                    raise RuntimeError(
                        "The molecule selection has 0 total mass. Please read atom masses from a prmtop or psf file."
                    )

        if "trajalnsel" in props:
            res["trajalnsel"] = None
            if self._trajalnsel is not None:
                res["trajalnsel"] = mol.atomselect(self._trajalnsel)
                if np.sum(res["trajalnsel"]) == 0:
                    raise RuntimeError("Alignment selection resulted in 0 atoms.")

        if "centersel" in props and self._pbc:
            res["centersel"] = mol.atomselect(self._centersel)
            if np.sum(res["centersel"]) == 0:
                raise RuntimeError("Centering selection resulted in 0 atoms.")
        return res

    def project(self, mol):
        """Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)

        mol = mol.copy()
        if self._pbc:
            mol.wrap(getMolProp("centersel"))

        trajalnsel = getMolProp("trajalnsel")

        if trajalnsel is not None:
            if self._refmol is None:
                mol.align(trajalnsel)
            else:
                mol.align(
                    trajalnsel, refmol=self._refmol, refsel=getMolProp("refalnsel")
                )

        atomsel = getMolProp("atomsel")
        masses = getMolProp("masses")
        total_mass = np.sum(masses)

        coords = mol.coords[atomsel].copy()

        # Calculate center of mass
        com = np.sum(coords * masses[:, None, None], axis=0) / total_mass

        # Calculate radius of gyration
        coords -= com  # Remove COM from the coordinates
        sq_coords = coords ** 2
        sq = np.sum(sq_coords, axis=1)
        sq_x = np.sum(sq_coords[:, [1, 2]], axis=1)
        sq_y = np.sum(sq_coords[:, [0, 2]], axis=1)
        sq_z = np.sum(sq_coords[:, [0, 1]], axis=1)
        sq_radius = np.array([sq, sq_x, sq_y, sq_z])

        rog = np.sqrt(np.sum(masses[None, :, None] * sq_radius, axis=1) / total_mass).T

        return rog

    def getMapping(self, mol):
        """Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        from pandas import DataFrame

        getMolProp = lambda prop: self._getMolProp(mol, prop)

        atomidx = np.where(getMolProp("atomsel"))[0]
        types = ["rog", "rog", "rog", "rog"]
        indexes = [atomidx, atomidx, atomidx, atomidx]
        description = [
            "Radius of gyration",
            "x component",
            "y component",
            "z component",
        ]
        return DataFrame(
            {"type": types, "atomIndexes": indexes, "description": description}
        )


import unittest


class _TestMetricCoordinate(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from os import path

        mol = Molecule(
            path.join(home(dataDir="test-projections"), "trajectory", "filtered.psf")
        )
        mol.read(path.join(home(dataDir="test-projections"), "trajectory", "traj.xtc"))

        self.mol = mol

    def test_project(self):
        metr = MetricGyration("protein")
        data = metr.project(self.mol)

        lastrog = np.array(
            [
                18.002188,
                17.97491,
                17.973713,
                17.949549,
                17.950699,
                17.9308,
                17.928637,
                18.000408,
                18.02674,
                17.98852,
                18.015263,
                17.934515,
                17.94321,
                17.949211,
                17.93479,
                17.924484,
                17.920536,
                17.860697,
                17.849443,
                17.879776,
            ],
            dtype=np.float32,
        )
        assert np.all(
            np.abs(data[-20:, 0] - lastrog) < 0.001
        ), "Radius of gyration calculation is broken"


if __name__ == "__main__":
    unittest.main(verbosity=2)


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#


def closestDistance(mol1, mol2):
    from moleculekit.projections.util import pp_calcMinDistances

    mol1 = mol1.copy()
    mol1.segid[:] = "X"
    mol2 = mol2.copy()
    mol2.segid[:] = "Y"

    mol1.append(mol2)

    sel1 = mol1.atomselect("segid X")[None, :]
    sel2 = mol1.atomselect("segid Y")[None, :]

    return pp_calcMinDistances(mol1, sel1, sel2, periodic=None).squeeze()


def isProteinProtonated(mol):
    numberHs = mol.atomselect("protein and hydrogen").sum()
    numberProt = mol.atomselect("protein").sum()
    numberHeavy = numberProt - numberHs

    # Rough heuristic of what is a significant number of hydrogens
    if numberHs < (numberHeavy * 0.7):
        return False
    return True


def isLigandOptimized(mol, num_planes=3):
    from itertools import combinations
    import numpy as np
    import random

    if mol.numAtoms < 4:
        # Can't tell easily if it's optimized or not
        return True

    def random_combination(iterable, r):
        "Random selection from itertools.combinations(iterable, r)"
        pool = tuple(iterable)
        n = len(pool)
        indices = sorted(random.sample(range(n), r))
        return tuple(pool[i] for i in indices)

    coords = mol.coords[:, :, 0]

    for _ in range(num_planes):
        atom_triplet = random_combination(range(mol.numAtoms), 3)
        atom_coords = coords[list(atom_triplet)]
        vec1 = atom_coords[1] - atom_coords[0]
        vec2 = atom_coords[2] - atom_coords[0]
        plane_normal = np.cross(vec1, vec2)
        plane_normal /= np.linalg.norm(plane_normal)

        all_vectors = coords - coords[atom_triplet[0]]
        projections = np.abs(np.dot(all_vectors, plane_normal))

        atoms_on_plane = np.where(projections < 1e-4)[0]
        # print(f"{atoms_on_plane} atoms lie on the plane defined by {atom_triplet}")
        # print(projections)

        # Arbitrary threshold. 3 Atoms will always be on the plane they define so I assume if two more are exactly on the same plane it's no chance
        if len(atoms_on_plane) > 5:
            return False

    return True


def isLigandDocked(prot, lig, threshold=10):
    dist = closestDistance(prot, lig)
    return dist < threshold


def areLigandsDocked(prot_file, sdf_file, threshold=10, max_check=None):
    from moleculekit.smallmol.smallmollib import SmallMolLib
    from moleculekit.molecule import Molecule

    not_docked = []
    prot = Molecule(prot_file)
    for i, lig in enumerate(SmallMolLib(sdf_file)):
        if max_check is not None and i >= max_check:
            break

        ligname = lig.ligname
        lig = lig.toMolecule()

        if not isLigandDocked(prot, lig, threshold):
            not_docked.append(ligname)

    return len(not_docked) == 0, not_docked


def areLigandsOptimized(sdf_file, num_planes=3, max_check=None):
    from moleculekit.smallmol.smallmollib import SmallMolLib

    not_optimized = []
    for i, lig in enumerate(SmallMolLib(sdf_file)):
        if max_check is not None and i >= max_check:
            break

        ligname = lig.ligname
        lig = lig.toMolecule()

        if not isLigandOptimized(lig, num_planes=num_planes):
            not_optimized.append(ligname)

    return len(not_optimized) == 0, not_optimized


def proteinHasBonds(mol):
    import numpy as np

    prot_idx = mol.atomselect("protein", indexes=True)
    num_prot_bonds = np.sum(np.all(np.isin(mol.bonds, prot_idx), axis=1))

    return num_prot_bonds >= len(prot_idx) - 1


import unittest


class _TestMoleculeChecks(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.molecule import Molecule
        from moleculekit.tools.preparation import proteinPrepare

        self.mol = Molecule("3ptb")

        self.prot = self.mol.copy()
        self.prot.filter("protein")

        self.prot_protonated = self.prot.copy()
        self.prot_protonated = proteinPrepare(self.prot_protonated)

        self.lig = self.mol.copy()
        self.lig.filter("resname BEN")

        self.lig_far = self.lig.copy()
        self.lig_far.moveBy([0, 20, 0])

        self.lig_flat = self.lig.copy()
        self.lig_flat.coords[:, 0, :] = -1.7  # Flatten on x axis

    def test_optimized_ligands(self):
        assert isLigandOptimized(self.lig)
        assert not isLigandOptimized(self.lig_flat)

    def test_docked_ligands(self):
        assert isLigandDocked(self.prot, self.lig)
        assert not isLigandDocked(self.prot, self.lig_far)

    def test_protonated_protein(self):
        assert not isProteinProtonated(self.prot)
        assert isProteinProtonated(self.prot_protonated)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# cython: cdivision=True

import numpy as np
cimport numpy as np
from libcpp.vector cimport vector
from libc.math cimport round, sqrt, acos

# We now need to fix a datatype for our arrays. I've used the variable
# DTYPE for this, which is assigned to the usual NumPy runtime
# type info object.
UINT32 = np.uint32
FLOAT32 = np.float32

# "ctypedef" assigns a corresponding compile-time type to DTYPE_t. For
# every type in the numpy module there's a corresponding compile-time
# type with a _t-suffix.
ctypedef np.uint32_t UINT32_t
ctypedef np.float32_t FLOAT32_t

import cython
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef void _calculate_ring_mean(
        int f,
        int start_idx, 
        int end_idx, 
        np.ndarray[UINT32_t, ndim=1] rings_atoms, 
        np.ndarray[FLOAT32_t, ndim=3] coords,
        np.ndarray[FLOAT32_t, ndim=1] ring_mean,
    ):
    cdef FLOAT32_t coor
    for i in range(3):
        ring_mean[i] = 0

    for rr in range(start_idx, end_idx):
        for i in range(3):
            coor = coords[rings_atoms[rr], i, f]
            ring_mean[i] = ring_mean[i] + coor

    for i in range(3):
        ring_mean[i] = ring_mean[i] / (end_idx - start_idx)


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef float _wrapped_dist(
        int f,
        np.ndarray[FLOAT32_t, ndim=1] point1,
        np.ndarray[FLOAT32_t, ndim=1] point2,
        np.ndarray[FLOAT32_t, ndim=1] half_box,
        np.ndarray[FLOAT32_t, ndim=2] box,
    ):
    cdef FLOAT32_t val
    cdef float dist2 = 0
    for i in range(3):
        val = point1[i] - point2[i] 
        # Wrap the distance vector into the periodic box
        if abs(val) > half_box[i] and box[i, f] != 0:
            val = val - box[i, f] * round(val / box[i, f])
        dist2 = dist2 + (val * val)
    return dist2


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef void _normalize(np.ndarray[FLOAT32_t, ndim=1] vec):
    cdef float vec_norm = 0 
    for i in range(3):
        vec_norm = vec_norm + (vec[i] * vec[i])
    vec_norm = sqrt(vec_norm)

    for i in range(3):
        vec[i] = vec[i] / vec_norm


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef void _cross_product(
        np.ndarray[FLOAT32_t, ndim=1] vec_a, 
        np.ndarray[FLOAT32_t, ndim=1] vec_b,
        np.ndarray[FLOAT32_t, ndim=1] res,
    ):
    res[0] = vec_a[1] * vec_b[2] - vec_a[2] * vec_b[1]
    res[1] = vec_a[2] * vec_b[0] - vec_a[0] * vec_b[2]
    res[2] = vec_a[0] * vec_b[1] - vec_a[1] * vec_b[0]

    _normalize(res)


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def calculate(
        np.ndarray[UINT32_t, ndim=1] rings_atoms, 
        np.ndarray[UINT32_t, ndim=1] rings_start_indexes, 
        np.ndarray[UINT32_t, ndim=1] cations,
        np.ndarray[FLOAT32_t, ndim=3] coords, 
        np.ndarray[FLOAT32_t, ndim=2] box,
        float dist_threshold=5,
        float angle_threshold_min=60,
    ):
    cdef int r1, r2, f, i
    cdef int n_rings = rings_start_indexes.shape[0] - 1 # Subtract 1 for the added 0 index at the start
    cdef int n_cations = cations.shape[0]
    cdef int n_frames = coords.shape[2]

    cdef vector[vector[int]] results
    cdef vector[vector[float]] distangles
    cdef np.ndarray[FLOAT32_t, ndim=1] ring_mean = np.zeros(3, dtype=FLOAT32)
    cdef np.ndarray[FLOAT32_t, ndim=1] ring_normal = np.zeros(3, dtype=FLOAT32)
    cdef np.ndarray[FLOAT32_t, ndim=1] half_box = np.zeros(3, dtype=FLOAT32)
    cdef np.ndarray[FLOAT32_t, ndim=1] cation_coor = np.zeros(3, dtype=FLOAT32)
    cdef np.ndarray[FLOAT32_t, ndim=1] tmp1 = np.zeros(3, dtype=FLOAT32)
    cdef np.ndarray[FLOAT32_t, ndim=1] tmp2 = np.zeros(3, dtype=FLOAT32)
    cdef FLOAT32_t dist2, dot_prod
    cdef int r_start_idx, r_end_idx

    dist_threshold = dist_threshold * dist_threshold

    for f in range(n_frames):
        results.push_back(vector[int]())
        distangles.push_back(vector[float]())
        for i in range(3):
            half_box[i] = box[i, f] / 2

        for r1 in range(n_rings):
            for r2 in range(n_cations):
                r_start_idx = rings_start_indexes[r1]
                r_end_idx = rings_start_indexes[r1+1]

                # Calculate the ring centroid
                _calculate_ring_mean(f, r_start_idx, r_end_idx, rings_atoms, coords, ring_mean)

                # Cation coordinates
                for i in range(3):
                    cation_coor[i] = coords[cations[r2], i, f]

                # Calculate the wrapped distance between the ring centroid and the cation
                dist2 = _wrapped_dist(f, ring_mean, cation_coor, half_box, box)

                # Ring centroid too far apart from cation
                if dist2 > dist_threshold:
                    continue

                # Calculate the plane normal
                for i in range(3):
                    tmp1[i] = coords[rings_atoms[r_start_idx], i, f] - coords[rings_atoms[r_start_idx+2], i, f]
                    tmp2[i] = coords[rings_atoms[r_start_idx+1], i, f] - coords[rings_atoms[r_start_idx+2], i, f]
                _cross_product(tmp1, tmp2, ring_normal)

                # Calculate ring-cation vector
                for i in range(3):
                    tmp1[i] = cation_coor[i] - ring_mean[i]
                _normalize(tmp1)

                # Calculate angle between plane normal and ring-cation vector
                dot_prod = 0
                for i in range(3):
                    dot_prod = dot_prod + ring_normal[i] * tmp1[i]
                angle = acos(dot_prod) * 57.29578  # Convert radians to degrees
                if angle > 90:  # minimal angle to line. necessary as perpendicular can point either way
                    angle = 180 - angle
                angle = 90 - angle # We want the angle to the plane, not the normal of the plane so 90 - angle
                
                if angle >= angle_threshold_min:
                    results[f].push_back(r1)
                    results[f].push_back(cations[r2])
                    distangles[f].push_back(sqrt(dist2))
                    distangles[f].push_back(angle)

    return results, distangles





try:
    import rdkit
except ImportError as e:
    raise ImportError(
        "{}. You are probably missing the rdkit package. Please install it to support rdkit integration "
        "features".format(e)
    )
from rdkit import Chem
from rdkit.Chem import ChemicalFeatures
from rdkit import RDConfig
from rdkit.Geometry import Point3D
import os
import numpy as np


def StandardPDBResidueChirality(rdmol):
    for a in rdmol.GetAtoms():
        if a.GetChiralTag() != rdkit.Chem.rdchem.CHI_UNSPECIFIED:
            info = a.GetMonomerInfo()
            if (
                info
                and (
                    info.GetMonomerType()
                    == rdkit.Chem.rdchem.AtomMonomerType.PDBRESIDUE
                )
                and not info.GetIsHeteroAtom()
                and not StandardPDBChiralAtom(info.GetResidueName(), info.GetName())
            ):
                a.SetChiralTag(rdkit.Chem.rdchem.CHI_UNSPECIFIED)
                # TODO: Missing CIPCode unsetting


def StandardPDBChiralAtom(resname, atomname):
    if resname == "GLY":
        return False
    if resname == "ILE" or resname == "THR":
        return atomname == "CA" or atomname == "CB"
    if resname in (
        "ALA",
        "ARG",
        "ASN",
        "ASP",
        "CYS",
        "GLN",
        "GLU",
        "HIS",
        "LEU",
        "LYS",
        "MET",
        "PHE",
        "PRO",
        "SER",
        "TRP",
        "TYR",
        "VAL",
    ):
        return atomname == "CA"
    return False


def toRDKITmol(mol, protidx, sanitize=True, removeHs=False):
    # Taken from rdkit/Code/GraphMol/FileParsers/PDBParser.cpp
    conformer = Chem.Conformer(len(protidx))
    conformer.Set3D(True)
    conformer.SetId(0)
    rdmol = Chem.RWMol()
    atomlist = []
    for ii, i in enumerate(protidx):
        a = Chem.Atom(mol.element[i])
        a.SetFormalCharge(int(mol.charge[i]))
        info = Chem.AtomPDBResidueInfo(
            atomName=mol.name[i],
            serialNumber=int(mol.serial[i]),
            altLoc=mol.altloc[i],
            residueName=mol.resname[i],
            residueNumber=int(mol.resid[i]),
            chainId=mol.chain[i],
            insertionCode=mol.insertion[i],
            occupancy=float(mol.occupancy[i]),
            tempFactor=float(mol.beta[i]),
            isHeteroAtom=mol.record[i] == "HETATM",
        )
        a.SetMonomerInfo(info)

        rdmol.AddAtom(a)
        atomlist.append(a)
        coor = [float(c) for c in mol.coords[i, :, mol.frame]]
        conformer.SetAtomPosition(
            ii, Point3D(coor[0], coor[1], coor[2])
        )  # Correct the atom idx
    rdmol.AddConformer(conformer)

    # Here I diverge from the C++ parser because you cannot instantiate Chem.Bond objects in python
    # I also don't take into account double/triple bonds etc since I don't think we actually store them in Molecule
    for b in mol._getBonds():
        if b[0] in protidx and b[1] in protidx:
            bond = rdmol.GetBondBetweenAtoms(int(b[0]), int(b[1]))
            if bond is None:
                rdmol.AddBond(
                    int(np.where(protidx == b[0])[0]),
                    int(np.where(protidx == b[1])[0]),
                    Chem.BondType.SINGLE,
                )

    # Proximitybonds I already did by using _getBonds which calls _guessBonds
    # TODO: Set PDB double bonds

    # Calculate explicit valence of atoms
    for a in atomlist:
        pass

    if sanitize:
        if removeHs:
            Chem.RemoveHs(rdmol)
        else:
            Chem.SanitizeMol(rdmol)
    else:
        rdmol.UpdatePropertyCache()

    # Set tetrahedral chirality from 3D co-ordinates
    Chem.AssignAtomChiralTagsFromStructure(rdmol)
    StandardPDBResidueChirality(rdmol)

    return rdmol


if __name__ == "__main__":
    from moleculekit.molecule import Molecule

    mol = Molecule("3PTB")
    mol.filter("resname BEN")
    res = _convertMoleculeToRDKitMol(mol)
    assert res is not None

    # # We don't use this yet. It's experimental. So no point in wasting time testing it.
    # protidx = mol.atomselect('protein', indexes=True)

    # rdmol = toRDKITmol(mol, protidx)

    # fdefName = os.path.join(RDConfig.RDDataDir,'BaseFeatures.fdef')
    # factory = ChemicalFeatures.BuildFeatureFactory(fdefName)
    # feats = factory.GetFeaturesForMol(rdmol)



#!/usr/bin/env python
# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import subprocess
import inspect
import threading
import platform
import queue
import time
import os
import sys
from moleculekit.support import string_to_tempfile
import numpy as np
import tempfile
import logging

logger = logging.getLogger(__name__)


_viewers = np.empty(0, dtype=object)


def _enqueue_output(obj, vmd, queue):
    while vmd.poll() is None:
        line = vmd.stdout.readline()
    # print("READIN: [" + line.decode("ascii") + "]", end="" );
    # 			queue.put(line.decode("ascii"))
    # 	print( "Process finished" )
    vmd.stdout.close()
    obj.done = True


class VMD:
    """Please do not directly call this class constructor. Use the `viewer` or `getCurrentViewer` function instead.

    Parameters
    ----------
    vmd : str
    host : str
    dispdev : str
    """

    def __init__(self, vmd=None, host=None, dispdev="win"):
        self.done = False
        vmd = getVMDpath(vmd=vmd)

        args = [vmd]
        if host:
            args.append("--host")
            args.append(host)
        args.append("--dispdev")
        args.append(dispdev)
        self.vmd = subprocess.Popen(
            args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            bufsize=0,
            close_fds=True,
            shell=False,
        )
        self.queue = queue.Queue()
        self.thread = threading.Thread(
            target=_enqueue_output, args=(self, self.vmd, self.queue)
        )
        self.thread.daemon = True
        self.thread.start()
        # 	self.vmd.stdin.write( b"chan configure stdout -buffering none\n" )
        time.sleep(2)
        self.send("display reposition 500 1000")
        self.send("display resize 800 800")
        self.send("menu main on")
        self.send("display depthcue off")
        self.send("axes location Off")
        # self.send('color Display Background white')
        self.send("display projection Orthographic")

    def send(self, command):
        """Send a tcl command to VMD

        Parameters
        ----------
        command : str
            The tcl command which to send
        """
        if self.done:
            return
        # print( command )
        fn = string_to_tempfile(command, "tcl")
        # 		print(fn)

        cc = f"if {{ [ catch {{ source {fn}}} ] }} {{ unlink {fn} }} else {{ unlink {fn} }}\n"
        self.vmd.stdin.write(cc.encode("ascii"))
        self.vmd.stdin.flush()

        while os.path.isfile(fn) and not self.done:
            time.sleep(0.01)

    def loadMol(self, mol, name=None):
        """Load a :class:`Molecule <moleculekit.molecule.Molecule>` object into VMD

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            The Molecule to load into VMD
        name : str
            The name to give to the Molecule
        """
        # Getting name of variable passed to method
        """
        caller = inspect.currentframe().f_back
        try:
            value = caller.f_locals[mol]
        except KeyError:
            value = caller.f_globals[mol]
        if name is None:
            name = value
        # -----------------------------------------
        """
        mol.view(name=name, viewerhandle=self)

    def rep(self, mode, sel="resname MOL", color=0):
        """Modify representations for the top molecule in VMD

        Parameters
        ----------
        mode : str ('protein', 'ligand')
            If set to 'protein', it will show a single conformation of the protein with sequence coloring
            If set to 'ligand', it will show all ligands in the specified `color` and the protein in teal
        sel : str
            Atom selection string for the ligand mode.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        color : int
            Color for the ligand. Use color numbers of VMD
        """
        if mode == "ligand":
            # Protein representation
            self.send("mol modstyle 0 top NewCartoon")
            self.send('mol modselect 0 top "protein"')
            # self.send('mol modcolor 0 top Index')
            # Ligand representation
            self.send("set rep [molinfo top get numreps]")
            self.send("mol color ColorID " + str(color))
            self.send("mol representation Lines 1.000000")
            self.send("mol selection " + sel)
            self.send("mol material Opaque")
            self.send("mol addrep top")
            self.send("set end [molinfo top get numframes]")
            self.send('mol drawframes top $rep "0:$end"')
        elif mode == "protein":
            # Protein representation
            self.send("mol modstyle 0 top NewCartoon")
            self.send('mol modselect 0 top "protein"')
            self.send("mol modcolor 0 top Index")
        else:
            raise ValueError(
                "Invalid mode. Choose between " "ligand" " and " "protein" ""
            )

    def completed(self):
        """Check if the viewer has been closed

        Returns
        -------
        comp : bool
            Returns True if the VMD viewer has closed
        """
        return self.done

    def copy(self):
        return None

    def render(
        self,
        outfile,
        renderer="TachyonInternal",
        resolution=None,
        aasamples=None,
        skylight=None,
        tachyon=None,
        convert=None,
        trim=False,
    ):
        """Renders the current VMD scene into a file.

        Parameters
        ----------
        outfile : str
            File to which to render image
        renderer : ('TachyonInternal', 'tachyon', 'snapshot')
            Which renderer to use
        resolution : tuple
            X,Y resolution of the output image i.e. (1920, 1080). Only used with the renderer='tachyon' option.
        aasamples : int
            Number of anti-aliasing samples. Only used with the renderer='tachyon' option.
        skylight : float
            Add a skylight. Only used with the renderer='tachyon' option.
        tachyon : str
            Path to tachyon renderer executable. Only used with the renderer='tachyon' option.
        convert : bool
            Attempts to convert the image to the datatype of the `outfile` extension
        trim : bool
            Trims the whitespace of the image
        """
        import shutil

        outfile = os.path.abspath(outfile)
        outname, ext = os.path.splitext(outfile)

        if (renderer.lower() != "tachyon") and (
            resolution is not None
            or aasamples is not None
            or skylight is not None
            or tachyon is not None
        ):
            raise AttributeError(
                "resolution, aasamples, skylight and tachyon parameters only accepted with the "
                "renderer='tachyon' option."
            )

        if renderer.lower() == "tachyon":
            tmpext = ".psd"
            if tachyon is None:
                tachyon = shutil.which("tachyon", mode=os.X_OK)
            if tachyon is None:
                raise FileNotFoundError(
                    "Could not find `tachyon` executable, or no execute permissions are given. Try using renderer='snapshot' instead."
                )
            rendercommand = f"render Tachyon {outname}"
        elif renderer.lower() == "snapshot":
            tmpext = ".tga"
            rendercommand = f"render snapshot {outname}{tmpext}"
        elif renderer.lower() == "tachyoninternal":
            tmpext = ".tga"
            rendercommand = f"render TachyonInternal {outname}{tmpext}"

        self.send(rendercommand)
        if renderer == "tachyon":
            os.system(
                f"{tachyon} -res {resolution[0]} {resolution[1]} -aasamples {aasamples} -add_skylight {skylight} {outname} -format PSD48 -o {outname}.psd"
            )
        logger.debug(rendercommand)
        if not os.path.exists(outname + tmpext):
            raise RuntimeError(
                f"Rendering failed to produce image with following command: {rendercommand}"
            )
        if os.path.exists(outname):
            os.remove(outname)

        if ext != tmpext:
            if convert is None:
                convert = shutil.which("convert", mode=os.X_OK)
            if convert is None:
                raise FileNotFoundError(
                    f"Could not find `convert` executable, or no execute permissions are given. You can find the temporary render file in {outname + tmpext}"
                )

            os.system(f"{convert} {outname}{tmpext} {outname}{ext}")
            if trim:
                os.system(f"{convert} {outname}{ext} -trim {outname}{ext}")
            os.remove(outname + tmpext)

    def close(self):
        self.send("exit")

    def __del__(self):
        self.close()


def getVMDpath(vmd=None):
    sys = platform.system()
    if not vmd:
        if sys == "Linux" or sys == "Darwin":
            vmd = os.path.join(os.path.dirname(inspect.getfile(VMD)), "vmd_wrapper")
        elif sys == "Windows":
            vmd = os.path.join(os.path.dirname(inspect.getfile(VMD)), "vmd_wrapper.bat")
        else:
            raise OSError("Don't know how to run VMD on platform [" + sys + "]")
    if (not vmd) or (not os.access(vmd, os.X_OK)):
        raise OSError("Cannot find VMD. Specify the location with 'vmd=' argument")
    return vmd


def getCurrentViewer(dispdev="win"):
    """Get the handle to the current molecular viewer

    Parameters
    ----------
    dispdev : str, ('win', 'text', 'cave', 'caveforms', 'none')
        Specify the type of graphical display to use. The possible display devices include:
        win: a standard graphics display window.
        text: do not provide any graphics display window.
        cave: use the CAVE virtual environment for display, forms are disabled.
        caveforms: use the CAVE virtual environment for display and with forms enabled. This is useful with -display machine:0 for remote display of the forms when the CAVE uses the local screen.

    Returns
    -------
    viewer : :class:`VMD` object
        Returns a handle to the current viewer
    """
    global _viewers
    # Check for dead viewers and remove them
    todrop = np.empty(0, dtype=int)
    for i in range(len(_viewers)):
        if _viewers[i].completed():
            todrop = np.append(todrop, i)
    _viewers = np.delete(_viewers, todrop)
    # Creating a new viewer if none exist or otherwise returning the last viewer
    if len(_viewers) == 0:
        _viewers = np.append(_viewers, VMD(dispdev=dispdev))
    return _viewers[-1]


def viewer(dispdev="win"):
    """Start a new molecular viewer

    Returns
    -------
    viewer : :class:`VMD` object
        Returns a handle to a new viewer
    """
    global _viewers
    # Check for dead viewers and remove them
    todrop = np.empty(0, dtype=int)
    for i in range(len(_viewers)):
        if _viewers[i].completed:
            todrop = np.append(todrop, i)
    _viewers = np.delete(_viewers, todrop)
    # Creating a new viewer
    handle = VMD(dispdev=dispdev)
    _viewers = np.append(_viewers, handle)
    return handle


def _tempfilename():
    return os.path.join(
        tempfile._get_default_tempdir(), next(tempfile._get_candidate_names())
    )


from unittest import TestCase


class _TestVMDViewer(TestCase):
    def test_send(self):
        viewer = getCurrentViewer(dispdev="text")
        viewer.send("menu main off")
        viewer.send("menu main on")
        viewer.close()

    def test_rep(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home

        mol = Molecule(os.path.join(home(dataDir="pdb"), "alanine.pdb"))

        viewer = getCurrentViewer(dispdev="text")
        viewer.loadMol(mol)
        viewer.rep(mode="protein", sel="protein")
        viewer.rep(mode="ligand", sel="protein")
        viewer.close()


if __name__ == "__main__":
    import unittest

    unittest.main(verbosity=2)


# GNU LGPL v3
# Copyright (C) 2015-2018 Acellera
# info@acellera.com

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import traceback
import sys


def getOpenBabelProperties(pdb, outfile):
    try:
        from openbabel import pybel
    except ImportError:
        print(
            "Could not import openbabel. The atomtyper requires this dependency so please install it with `conda install openbabel -c conda-forge`"
        )
        sys.exit(1)

    try:
        mpybel = next(pybel.readfile("pdb", pdb))
    except Exception:
        traceback.print_exc()
        sys.exit(2)

    try:
        with open(outfile, "w") as f:
            for r in pybel.ob.OBResidueIter(mpybel.OBMol):
                for at in pybel.ob.OBResidueAtomIter(r):
                    f.write(
                        f"{at.GetIndex()},{r.GetName()},{r.GetNum()},{r.GetAtomID(at)},{at.GetType()},{at.GetPartialCharge():.3f}\n"
                    )
    except Exception:
        traceback.print_exc()
        sys.exit(3)


if __name__ == "__main__":
    getOpenBabelProperties(sys.argv[1], sys.argv[2])



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import sys


def getOpenBabelProperties(mol):
    import subprocess
    import os
    from moleculekit.util import tempname

    pdbfile = tempname(suffix=".pdb")
    outfile = tempname(suffix=".csv")
    mol.write(pdbfile)

    obabelcli = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "obabel_cli.py"
    )

    try:
        output = subprocess.check_output(
            [sys.executable, obabelcli, pdbfile, outfile],
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
    except subprocess.CalledProcessError as exc:
        print(
            "Failed to call getOpenBabelProperties with error",
            exc.returncode,
            exc.output,
        )
    else:
        print(output)

    os.remove(pdbfile)

    atoms = []
    with open(outfile, "r") as f:
        for line in f:
            pieces = line.split(",")
            pieces[0] = int(pieces[0])
            pieces[2] = int(pieces[2])
            pieces[5] = float(pieces[5])
            atoms.append(pieces)

    atoms.sort(key=lambda x: x[0])  # Sort by index

    return atoms


def openbabelConvert(input_file, input_format, output_format, extra_args=()):
    """
    Converts the file from the input format to the output format specified. It uses the openbabel features

    Parameters
    ----------
    input_file: str
        The path of the input file to convert
    input_format: str
        The input file format
    output_format: str
        The output file format

    Returns
    -------
    outfile: str
        The output file generated
    """
    import subprocess
    from moleculekit.util import tempname

    input_format = input_format[1:] if input_format.startswith(".") else input_format
    output_format = (
        output_format[1:] if output_format.startswith(".") else output_format
    )

    outfile = tempname(suffix=f".{output_format}")
    try:
        output = subprocess.check_output(
            [
                "obabel",
                f"-i{input_format}",
                input_file,
                f"-o{output_format}",
                f"-O{outfile}",
                *extra_args,
            ],
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
    except subprocess.CalledProcessError as exc:
        print("Failed to call openbabel with error", exc.returncode, exc.output)
    else:
        print(output)

    return outfile



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import logging
import tempfile
import numpy as np
import os
from moleculekit.tools.preparationdata import PreparationData
from moleculekit.molecule import Molecule
try:
    from pdb2pqr.main import runPDB2PQR
    from pdb2pqr.src.pdb import readPDB
except ImportError:
    raise ImportError('pdb2pqr not installed. To use the molecule preparation features please do `conda install htmd-pdb2pqr -c acellera`')

try:
    import propka.lib
except ImportError:
    raise ImportError('propka package not installed. To use the molecule preparation features please do `conda install propka -c acellera`')


logger = logging.getLogger(__name__)


def _selToHoldList(mol, sel):
    if sel:
        tx = mol.copy()
        tx.filter(sel)
        tx.filter("name CA")
        ret = list(zip(tx.resid, tx.chain, tx.insertion))
    else:
        ret = None
    return ret


def _fillMolecule(name, resname, chain, resid, insertion, coords, segid, element,
                  occupancy, beta, charge, record):
    numAtoms = len(name)
    mol = Molecule()
    mol.empty(numAtoms)

    mol.name = np.array(name, dtype=mol._dtypes['name'])
    mol.resname = np.array(resname, dtype=mol._dtypes['resname'])
    mol.chain = np.array(chain, dtype=mol._dtypes['chain'])
    mol.resid = np.array(resid, dtype=mol._dtypes['resid'])
    mol.insertion = np.array(insertion, dtype=mol._dtypes['insertion'])
    mol.coords = np.array(np.atleast_3d(np.vstack(coords)), dtype=mol._dtypes['coords'])
    mol.segid = np.array(segid, dtype=mol._dtypes['segid'])
    mol.element = np.array(element, dtype=mol._dtypes['element'])
    mol.occupancy = np.array(occupancy, dtype=mol._dtypes['occupancy'])
    mol.beta = np.array(beta, dtype=mol._dtypes['beta'])
    mol.box = np.zeros((3, mol.coords.shape[2]), dtype=mol._dtypes['box'])
    # mol.charge = np.array(charge, dtype=mol._dtypes['charge'])
    # mol.record = np.array(record, dtype=mol._dtypes['record'])
    return mol


def _fixupWaterNames(mol):
    """Rename WAT / OW HW HW atoms as O H1 H2"""
    mol.set("name", "O", sel="resname WAT and name OW")
    mol.set("name", "H1", sel="resname WAT and name HW and serial % 2 == 0")
    mol.set("name", "H2", sel="resname WAT and name HW and serial % 2 == 1")


def _warnIfContainsDUM(mol):
    """Warn if any DUM atom is there"""
    if any(mol.atomselect("resname DUM")):
        logger.warning("OPM's DUM residues must be filtered out before preparation. Continuing, but crash likely.")

def _checkChainAndSegid(mol, _loggerLevel):
    from moleculekit.util import sequenceID
    import string
    emptychains = mol.chain == ''
    emptysegids = mol.segid == ''

    if np.all(emptychains) and np.all(emptysegids):
        raise RuntimeError('No chains or segments defined in Molecule.chain / Molecule.segid. Please assign either to continue with preparation.')

    if np.all(emptychains) and np.any(~emptysegids):
        logger.info('No chains defined in Molecule. Using segment IDs as chains for protein preparation.')
        mol = mol.copy()
        chain_alphabet = np.array(list(string.digits + string.ascii_uppercase + string.ascii_lowercase))
        mol.chain[:] = chain_alphabet[sequenceID(mol.segid)]
        
    if np.any(~emptysegids) and np.any(~emptychains):
        chainseq = sequenceID(mol.chain)
        segidseq = sequenceID(mol.segid)
        if not np.array_equal(chainseq, segidseq):
            logger.warning('Both chains and segments are defined in Molecule.chain / Molecule.segid, however they are inconsistent. ' \
                           'Protein preparation will use the chain information.')

    if _loggerLevel is None or _loggerLevel == 'INFO':
        chainids = np.unique(mol.chain)
        if np.any([len(cc) > 1 for cc in chainids]):
            raise RuntimeError('The chain field should only contain a single character.')
            
        print('\n---- Molecule chain report ----')
        for c in chainids:
            chainatoms = np.where(mol.chain == c)[0]
            firstatom = chainatoms[0]
            lastatom = chainatoms[-1]
            print(f'Chain {c}:')
            print(f'    First residue: {mol.resname[firstatom]}:{mol.resid[firstatom]}:{mol.insertion[firstatom]}')
            print(f'    Final residue: {mol.resname[lastatom]}:{mol.resid[lastatom]}:{mol.insertion[lastatom]}')
        print('---- End of chain report ----\n')
    
    return mol


def _buildResAndMol(pdb2pqr_protein):
    # Here I parse the returned protein object and recreate a Molecule,
    # because I need to access the properties.
    logger.debug("Building Molecule object.")

    name = []
    resid = []
    chain = []
    insertion = []
    coords = []
    resname = []
    segid = []
    element = []
    occupancy = []
    beta = []
    record = []
    charge = []

    prepData = PreparationData()

    for i, residue in enumerate(pdb2pqr_protein.residues):
        # if 'ffname' in residue.__dict__:
        if getattr(residue, 'ffname', None):
            curr_resname = residue.ffname
            if len(curr_resname) >= 4:
                curr_resname = curr_resname[-3:]
                logger.debug("Residue %s has internal name %s, replacing with %s" %
                             (residue, residue.ffname, curr_resname))
        else:
            curr_resname = residue.name

        prepData._setProtonationState(residue, curr_resname)

        # Removed because not really useful
        # if getattr(residue, 'patches', None):
        #     for patch in residue.patches:
        #         prepData._appendPatches(residue, patch)
        #         if patch != "PEPTIDE":
        #             logger.debug("Residue %s has patch %s set" % (residue, patch))

        if getattr(residue, 'wasFlipped', 'UNDEF') != 'UNDEF':
            prepData._setFlipped(residue, residue.wasFlipped)

        prepData._set(residue, 'pdb2pqr_idx', i)

        for atom in residue.atoms:
            name.append(atom.name)
            resid.append(residue.resSeq)
            chain.append(residue.chainID)
            insertion.append(residue.iCode)
            coords.append([atom.x, atom.y, atom.z])
            resname.append(curr_resname)
            segid.append(atom.segID)
            # Fixup element fields for added H (routines.addHydrogens)
            elt = "H" if atom.added and atom.name.startswith("H") else atom.element
            element.append(elt.capitalize())
            occupancy.append(0.0 if atom.added else atom.occupancy)
            beta.append(99.0 if atom.added else atom.tempFactor)
            charge.append(atom.charge)
            record.append(atom.type)
            if atom.added:
                logger.debug("Coordinates of atom {:s} in residue {:s} were guessed".format(residue.__str__(),atom.name))
                prepData._set(residue, 'guessedAtoms', atom.name, append=True)


    mol_out = _fillMolecule(name, resname, chain, resid, insertion, coords, segid, element,
                            occupancy, beta, charge, record)
    # mol_out.set("element", " ")
    # Re-calculating elements
    # mol_out.element[:] = ''
    assert not np.any(mol_out.element == 'X'), "pdb2pqr left some lonepair atoms in. report this issue to moleculekit github issue tracker"
    mol_out.element = mol_out._guessMissingElements()

    prepData._importPKAs(pdb2pqr_protein.pka_protein)

    return mol_out, prepData


def proteinPrepare(mol_in,
                   pH=7.0,
                   verbose=0,
                   returnDetails=False,
                   hydrophobicThickness=None,
                   holdSelection=None,
                   _loggerLevel=None):
    """A system preparation wizard for HTMD.

    Returns a Molecule object, where residues have been renamed to follow
    internal conventions on protonation (below). Coordinates are changed to
    optimize the H-bonding network. This should be roughly equivalent to mdweb and Maestro's
    preparation wizard.

    The following residue names are used in the returned molecule:

    === ===============================
    ASH Neutral ASP
    CYX SS-bonded CYS
    CYM Negative CYS
    GLH Neutral GLU
    HIP Positive HIS
    HID Neutral HIS, proton HD1 present
    HIE Neutral HIS, proton HE2 present
    LYN Neutral LYS
    TYM Negative TYR
    AR0 Neutral ARG
    === ===============================

    ========= ======= =========
    Charge +1 Neutral Charge -1
    ========= ======= =========
    -         ASH     ASP
    -         CYS     CYM
    -         GLH     GLU
    HIP       HID/HIE -
    LYS       LYN     -
    -         TYR     TYM
    ARG       AR0     -
    ========= ======= =========

    A detailed table about the residues modified is returned (as a second return value) when
    returnDetails is True (see PreparationData object).

    If hydrophobicThickness is set to a positive value 2*h, a warning is produced for titratable residues
    having -h<z<h and are buried in the protein by less than 75%. Note that the heuristic for the
    detection of membrane-exposed residues is very crude; the "buried fraction" computation
    (from propKa) is approximate; also, in the presence of cavities,
    residues may be solvent-exposed independently from their z location.


    Notes
    -----
    In case of problems, exclude water and other dummy atoms.

    Features:
     - assign protonation states via propKa
     - flip residues to optimize H-bonding network
     - debump collisions
     - fill-in missing atoms, e.g. hydrogen atoms


    Parameters
    ----------
    mol_in : moleculekit.molecule.Molecule
        the object to be optimized
    pH : float
        pH to decide titration
    verbose : int
        verbosity
    returnDetails : bool
        whether to return just the prepared Molecule (False, default) or a molecule *and* a ResidueInfo
        object including computed properties
    hydrophobicThickness : float
        the thickness of the membrane in which the protein is embedded, or None if globular protein.
        Used to provide a warning about membrane-exposed residues.
    holdSelection : str
        (Untested) Atom selection to be excluded from optimization.
        Only the carbon-alpha atom will be considered for the corresponding residue.


    Returns
    -------
    mol_out : Molecule
        the molecule titrated and optimized. The molecule object contains an additional attribute,
    resData : PreparationData
        a table of residues with the corresponding protonation states, pKas, and other information


    Examples
    --------
    >>> tryp = Molecule('3PTB')

    >>> tryp_op, prepData = proteinPrepare(tryp, returnDetails=True)
    >>> tryp_op.write('proteinpreparation-test-main-ph-7.pdb')
    >>> prepData.data.to_excel("/tmp/tryp-report.xlsx")
    >>> prepData                                                        # doctest: +NORMALIZE_WHITESPACE
    PreparationData object about 290 residues.
    Unparametrized residue names: CA, BEN
    Please find the full info in the .data property, e.g.:
      resname  resid insertion chain       pKa protonation flipped     buried
    0     ILE     16               A       NaN         ILE     NaN        NaN
    1     VAL     17               A       NaN         VAL     NaN        NaN
    2     GLY     18               A       NaN         GLY     NaN        NaN
    3     GLY     19               A       NaN         GLY     NaN        NaN
    4     TYR     20               A  9.590845         TYR     NaN  14.642857
     . . .
    >>> x_HIE91_ND1 = tryp_op.get("coords","resid 91 and  name ND1")
    >>> x_SER93_H =   tryp_op.get("coords","resid 93 and  name H")
    >>> len(x_SER93_H) == 3
    True
    >>> np.linalg.norm(x_HIE91_ND1-x_SER93_H) < 3
    True

    >>> tryp_op = proteinPrepare(tryp, pH=1.0)
    >>> tryp_op.write('proteinpreparation-test-main-ph-1.pdb')

    >>> tryp_op = proteinPrepare(tryp, pH=14.0)
    >>> tryp_op.write('proteinpreparation-test-main-ph-14.pdb')

    >>> mol = Molecule("1r1j")
    >>> mo, prepData = proteinPrepare(mol, returnDetails=True)
    >>> prepData.missedLigands
    ['NAG', 'ZN', 'OIR']

    >>> his = prepData.data.resname == "HIS"
    >>> prepData.data[his][["resid","insertion","chain","resname","protonation"]]
         resid insertion chain resname protonation
    160    214               A     HIS         HID
    163    217               A     HIS         HID
    383    437               A     HIS         HID
    529    583               A     HIS         HID
    533    587               A     HIS         HIP
    583    637               A     HIS         HID
    627    681               A     HIS         HID
    657    711               A     HIS         HIP
    679    733               A     HIS         HID

    >>> mor = Molecule("4dkl")
    >>> mor.filter("protein and noh")
    >>> mor_opt, mor_data = proteinPrepare(mor, returnDetails=True,
    ...                                    hydrophobicThickness=32.0)
    >>> exposedRes = mor_data.data.membraneExposed
    >>> mor_data.data[exposedRes].to_excel("/tmp/mor_exposed_residues.xlsx")

    >>> im=Molecule("4bkj")
    >>> imo,imd=proteinPrepare(im,returnDetails=True)
    >>> imd.data.to_excel("/tmp/imatinib_report.xlsx")

    See Also
    --------
    :class:`moleculekit.tools.preparationdata.PreparationData`

    Notes
    -----
    Unsupported/To Do/To Check:
     - ligands
     - termini
     - multiple chains
     - nucleic acids
     - coupled titrating residues
     - Disulfide bridge detection (implemented but unused)
    """
    from moleculekit.tools.preparationdata import logger as datalogger
    oldLoggingLevel = [logger.level, datalogger.level]
    if verbose:
        logger.setLevel(logging.DEBUG)
        datalogger.setLevel(logging.DEBUG)
    if _loggerLevel is not None:
        logger.setLevel(_loggerLevel)
        datalogger.setLevel(_loggerLevel)
    logger.debug("Starting.")

    _warnIfContainsDUM(mol_in)

    mol_in = _checkChainAndSegid(mol_in, _loggerLevel)

    # We could transform the molecule into an internal object, but for
    # now I prefer to rely on the strange internal parser to avoid
    # hidden quirks.
    tmpin = tempfile.NamedTemporaryFile(suffix=".pdb", mode="w+")
    logger.debug("Temporary file is " + tmpin.name)
    mol_in.write(tmpin.name)  # Not sure this is good unix

    pdblist, errlist = readPDB(tmpin)
    if len(pdblist) == 0 and len(errlist) == 0:
        raise Exception('Internal error in preparing input to pdb2pqr')

    # An ugly hack to silence non-prefixed logging messages
    for h in propka.lib.logger.handlers:
        if h.formatter._fmt == '%(message)s':
            propka.lib.logger.removeHandler(h)

    propka_opts, dummy = propka.lib.loadOptions('--quiet')
    propka_opts.verbosity = verbose
    propka_opts.verbose = verbose  # Will be removed in future propKas

    # Note on naming. The behavior of PDB2PQR is controlled by two
    # parameters, ff and ffout. My understanding is that the ff
    # parameter sets which residues are SUPPORTED by the underlying
    # FF, PLUS the charge and radii.  The ffout parameter sets the
    # naming scheme. Therefore, I want ff to be as general as
    # possible, which turns out to be "parse". Then I pick a
    # convenient ffout.

    # Hold list (None -> None)
    hlist = _selToHoldList(mol_in, holdSelection)
    if hlist:
        logger.warning("The holdSelection option is untested and deprecated. Please use reprepare()")

    # Relying on defaults
    pqr_res = runPDB2PQR(pdblist,
                         ph=pH, verbose=verbose,
                         ff="parse", ffout="amber",
                         ph_calc_method="propka31",
                         ph_calc_options=propka_opts,
                         holdList=hlist)
    try:
        header, pqr, missedLigands, pdb2pqr_protein, pdb2pqr_routines = \
            pqr_res['header'], pqr_res['lines'], pqr_res['missedligands'], pqr_res['protein'], pqr_res['routines']
    except:
        logger.error("Problem calling pdb2pqr. Make sure you have htmd-pdb2pqr >= 2.1.2a9")
        raise

    tmpin.close()

    # Diagnostics
    for missedligand in missedLigands:
        logger.warning("The following residue has not been optimized: " + missedligand)

    mol_out, resData = _buildResAndMol(pdb2pqr_protein)

    mol_out.box = mol_in.box
    _fixupWaterNames(mol_out)

    # Misc. info
    resData.header = header
    resData.pqr = pqr

    # Return residue information
    resData.pdb2pqr_protein = pdb2pqr_protein
    resData.pdb2pqr_routines = pdb2pqr_routines
    resData.missedLigands = missedLigands

    # Store un-reprepared info
    resData.data['default_protonation'] = resData.data['protonation']

    resData._listNonStandardResidues()
    resData._warnIfpKCloseTopH(pH)
    resData.warnIfTerminiSuspect()

    if hydrophobicThickness:
        resData._setMembraneExposureAndWarn(hydrophobicThickness)

    logger.debug("Returning.")
    logger.setLevel(oldLoggingLevel[0])
    datalogger.setLevel(oldLoggingLevel[1])

    if returnDetails:
        return mol_out, resData
    else:
        return mol_out


import unittest
class _TestPreparation(unittest.TestCase):
    def test_proteinPrepare(self):
        tryp_op, prepData = proteinPrepare(Molecule("3PTB"), returnDetails=True)
        d = prepData.data
        assert d.protonation[d.resid == 40].iloc[0] == 'HIE'
        assert d.protonation[d.resid == 57].iloc[0] == 'HIP'
        assert d.protonation[d.resid == 91].iloc[0] == 'HID'
        
    @unittest.skipUnless(os.environ.get('HTMD_LONGTESTS') == 'yes', 'Too long')
    def test_proteinPrepareLong(self):
        from moleculekit.home import home
        from moleculekit.util import assertSameAsReferenceDir
        
        pdbids = ['3PTB', '1A25', '1GZM', '1U5U']
        for pdb in pdbids:
            mol = Molecule(pdb)
            mol.filter("protein")
            mol_op, prepData = proteinPrepare(mol, returnDetails=True)
            mol_op.write("./{}-prepared.pdb".format(pdb))
            prepData.data.to_csv("./{}-prepared.csv".format(pdb), float_format="%.2f")
            compareDir = home(dataDir=os.path.join('test-proteinprepare', pdb))
            assertSameAsReferenceDir(compareDir)


if __name__ == "__main__":
    unittest.main(verbosity=2)
    import doctest
    doctest.testmod()


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

from moleculekit.projections.projection import Projection
import numpy as np
import logging

logger = logging.getLogger(__name__)


class MetricSecondaryStructure(Projection):
    """Calculates the secondary structure of the protein. DSSP implementation and documentation taken from MDtraj.

    Parameters
    ----------
    sel : str
        Atom selection string for the protein.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    simplified: bool
        Uses the simplified 3-letter code
    integer : bool
        Use integers instead of letter codes.

    Notes
    -----
    The simplified DSSP codes are:
        - 'C' : Coil. Either of the 'T', 'S' or ' ' codes. Integer code: 0
        - 'E' : Strand. Either of the 'E', or 'B' codes. Integer code: 1
        - 'H' : Helix. Either of the 'H', 'G', or 'I' codes. Integer code: 2

    The full DSSP assignment codes are:
        - 'H' : Alpha helix. Integer code: 3
        - 'B' : Residue in isolated beta-bridge. Integer code: 4
        - 'E' : Extended strand, participates in beta ladder. Integer code: 5
        - 'G' : 3-helix (3/10 helix). Integer code: 6
        - 'I' : 5 helix (pi helix). Integer code: 7
        - 'T' : hydrogen bonded turn. Integer code: 8
        - 'S' : bend. Integer code: 9
        - ' ' : Loops and irregular elements. Integer code: 10

    A special 'NA' code will be assigned to each 'residue' in the topology which
    isn't actually a protein residue (does not contain atoms with the names
    'CA', 'N', 'C', 'O'), such as water molecules that are listed as 'residue's
    in the topology.
    """

    def __init__(self, sel="protein", simplified=True, integer=True):
        super().__init__()

        self.sel = sel
        self.simplified = simplified
        self.integer = integer

    def _precalculate(self, mol):
        (
            self._pc_ca_indices,
            self._pc_nco_indices,
            self._pc_proline_indices,
            self._pc_chain_ids,
        ) = self._calcarrays(mol)

    def _calculateMolProp(self, mol, props="all"):
        from moleculekit.util import sequenceID

        # Calculate all properties at once since it would be too slow otherwise to redo calculations
        res = {}

        mol = mol.copy()
        mol.filter(self.sel, _logger=False)

        residues = sequenceID((mol.resid, mol.chain, mol.insertion))

        backbone = mol.atomselect("backbone")
        ca_indices = np.where(mol.name == "CA")[0].astype(np.int32)
        chainids = mol.chain[ca_indices]
        resnames = mol.resname[ca_indices]
        proline_indices = np.array(resnames == "PRO", dtype=np.int32)

        _, chain_ids = np.unique(chainids, return_inverse=True)
        chain_ids = chain_ids.astype(np.int32)

        nco_indices = np.ones((residues.max() + 1, 3), dtype=np.int32) * -1
        natriums = np.where((mol.name == "N") & backbone)[0]
        carbons = np.where((mol.name == "C") & backbone)[0]
        oxygens = np.where((mol.name == "O") & backbone)[0]
        nco_indices[residues[natriums], 0] = natriums
        nco_indices[residues[carbons], 1] = carbons
        nco_indices[residues[oxygens], 2] = oxygens

        res["ca_indices"] = ca_indices
        res["nco_indices"] = nco_indices
        res["proline_indices"] = proline_indices
        res["chain_ids"] = chain_ids
        return res

    def project(self, mol):
        """Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        mol = mol.copy()
        molprops = self._getMolProp(mol, "all")

        ca_indices = molprops["ca_indices"]
        nco_indices = molprops["nco_indices"]
        proline_indices = molprops["proline_indices"]
        chain_ids = molprops["chain_ids"]

        mol.filter(self.sel, _logger=False)

        xyz = np.swapaxes(np.swapaxes(np.atleast_3d(mol.coords), 1, 2), 0, 1)
        xyz = np.array(xyz.copy(), dtype=np.float32) / 10  # converting to nm

        try:
            from mdtraj.geometry._geometry import _dssp as dssp
        except ImportError:
            raise ImportError(
                "To calculate secondary structure you need to install mdtraj with `conda install mdtraj -c conda-forge`"
            )

        natoms = np.unique(
            (
                nco_indices.shape[0],
                ca_indices.shape[0],
                proline_indices.shape[0],
                chain_ids.shape[0],
            )
        )
        if len(natoms) != 1:
            raise AssertionError(
                "Wrong dimensions in SS data arrays. Report this bug on the issue tracker."
            )
        data = dssp(xyz, nco_indices, ca_indices, proline_indices, chain_ids)

        if self.simplified:
            trans = str.maketrans("HGIEBTS ", "HHHEECCC")
            data = data.translate(trans)

        data = np.fromiter(data, dtype=np.dtype("U2"))
        data = data.reshape(mol.numFrames, len(chain_ids))

        if self.integer:
            if self.simplified:
                data[data == "H"] = 2
                data[data == "E"] = 1
                data[data == "C"] = 0
            else:
                data[data == "H"] = 3
                data[data == "B"] = 4
                data[data == "E"] = 5
                data[data == "G"] = 6
                data[data == "I"] = 7
                data[data == "T"] = 8
                data[data == "S"] = 9
                data[data == " "] = 10
            data = np.array(data, dtype=np.int32)
        # data[:, np.logical_not(protein_indices)] = 'NA'

        return data

    def getMapping(self, mol):
        """Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        idx = mol.atomselect("{} and name CA".format(self.sel), indexes=True)
        from pandas import DataFrame

        types = []
        indexes = []
        description = []
        for i in idx:
            types += ["secondary structure"]
            indexes += [i]
            description += [
                "Secondary structure of residue {} {}{}".format(
                    mol.resname[i], mol.resid[i], mol.insertion[i]
                )
            ]
        return DataFrame(
            {"type": types, "atomIndexes": indexes, "description": description}
        )


def _ssmap(sschar):
    # Map secondary structures to integers
    #     G = 3-turn helix (310 helix). Min length 3 residues.
    #     H = 4-turn helix (α helix). Min length 4 residues.
    #     I = 5-turn helix (π helix). Min length 5 residues.
    #     T = hydrogen bonded turn (3, 4 or 5 turn)
    #     E = extended strand in parallel and/or anti-parallel β-sheet conformation. Min length 2 residues.
    #     B = residue in isolated β-bridge (single pair β-sheet hydrogen bond formation)
    #     S = bend (the only non-hydrogen-bond based assignment).
    #     C = coil (residues which are not in any of the above conformations).
    ssnum = np.zeros((len(sschar), len(sschar[0])), dtype=np.uint8)

    for i in range(len(sschar)):
        sschar[i][sschar[i] == "G"] = "H"
        sschar[i][sschar[i] == "I"] = "H"
        sschar[i][sschar[i] == "B"] = "E"
        sschar[i][sschar[i] == "b"] = "E"
        sschar[i][sschar[i] == "S"] = "C"
        sschar[i][sschar[i] == "T"] = "C"
        ssnum[i, np.where(sschar[i] == "C")[0]] = 0
        ssnum[i, np.where(sschar[i] == "E")[0]] = 1
        ssnum[i, np.where(sschar[i] == "H")[0]] = 2
    return ssnum


# fmt: off
import unittest
class _TestMetricSecondaryStructure(unittest.TestCase):
    def test_2hbb(self):
        from moleculekit.molecule import Molecule
        import numpy as np

        mol = Molecule('2HBB')  # NTL9
        mol.filter('protein')
        metr = MetricSecondaryStructure()
        data = metr.project(mol)
        ref = np.array([[0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0,
                        2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 2, 2, 2, 2,
                        2, 2, 2, 2, 2, 2, 0]], dtype=np.int32)
        assert np.array_equal(data, ref), 'MetricSecondaryStructure assertion failed'

        metr = MetricSecondaryStructure('resid 5 to 49')
        data = metr.project(mol)
        ref = np.array([[0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2,
                        2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 0]], dtype=np.int32)
        assert np.array_equal(data, ref), 'MetricSecondaryStructure assertion failed'

    def test_3ptb(self):
        from moleculekit.molecule import Molecule
        import numpy as np

        mol = Molecule('3PTB')  # Contains insertion which used to be a problem
        mol.filter('protein')
        metr = MetricSecondaryStructure()
        data = metr.project(mol)
        ref = np.array([[0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0,
                        1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 2, 2, 2, 0, 0, 0,
                        0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1,
                        1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1,
                        1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0,
                        0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                        0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0,
                        0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
                        0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
                        0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
                        2, 0, 0]], dtype=np.int32)
        assert np.array_equal(data, ref), 'MetricSecondaryStructure assertion failed'
        assert MetricSecondaryStructure().getMapping(mol).shape == (223, 3)

# fmt: on
if __name__ == "__main__":
    unittest.main(verbosity=2)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import string
import numpy as np
import logging


logger = logging.getLogger(__name__)


chain_alphabet = list(string.digits + string.ascii_uppercase + string.ascii_lowercase)
segid_alphabet = list(string.ascii_uppercase + string.ascii_lowercase + string.digits)


def autoSegment(
    mol,
    sel="all",
    basename="P",
    spatial=True,
    spatialgap=4.0,
    fields=("segid",),
    field=None,
    _logger=True,
):
    """Detects resid gaps in a selection and assigns incrementing segid to each fragment

    !!!WARNING!!! If you want to use atom selections like 'protein' or 'fragment',
    use this function on a Molecule containing only protein atoms, otherwise the protein selection can fail.

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule object
    sel : str
        Atom selection string on which to check for gaps.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    basename : str
        The basename for segment ids. For example if given 'P' it will name the segments 'P1', 'P2', ...
    spatial : bool
        Only considers a discontinuity in resid as a gap if the CA atoms have distance more than `spatialgap` Angstrom
    spatialgap : float
        The size of a spatial gap which validates a discontinuity (A)
    fields : list
        Fields in which to set the segments. Must be a combination of "chain", "segid" or only one of them.

    Returns
    -------
    newmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A new Molecule object with modified segids

    Example
    -------
    >>> newmol = autoSegment(mol, "chain B", "P", fields=("chain", "segid"))
    """
    from moleculekit.util import sequenceID

    if field is not None and isinstance(field, str):
        if field == "both":
            fields = ("chain", "segid")
        else:
            fields = (field,)

    mol = mol.copy()

    idx = mol.atomselect(sel, indexes=True)
    rid = mol.resid[idx].copy()
    residiff = np.diff(rid)
    # Points to the index before the gap!
    gappos = np.where((residiff != 1) & (residiff != 0))[0]

    idxstartseg = [idx[0]] + idx[gappos + 1].tolist()
    idxendseg = idx[gappos].tolist() + [idx[-1]]

    # Letters to be used for chains, if free: 0123456789abcd...ABCD..., minus chain symbols already used
    sel_mask = mol.atomselect(sel)
    used_chains = set(mol.chain[~sel_mask])
    available_chains = [x for x in chain_alphabet if x not in used_chains]
    used_segids = set([x[0] for x in mol.segid[~sel_mask] if x != ""])
    available_segids = [x for x in [basename] + segid_alphabet if x not in used_segids]
    basename = available_segids[0]

    if len(gappos) == 0:
        if "chain" in fields:
            mol.set("chain", available_chains[0], sel)
        if "segid" in fields:
            mol.set("segid", basename + "0", sel)
        return mol

    if spatial:
        residbackup = mol.resid.copy()
        # Assigning unique resids to be able to do the distance selection
        mol.set("resid", sequenceID(mol.resid))

        todelete = []
        i = 0
        for s, e in zip(idxstartseg[1:], idxendseg[:-1]):
            # Get the carbon alphas of both residues  ('coords', sel='resid "{}" "{}" and name CA'.format(mol.resid[e], mol.resid[s]))
            ca1coor = mol.coords[(mol.resid == mol.resid[e]) & (mol.name == "CA")]
            ca2coor = mol.coords[(mol.resid == mol.resid[s]) & (mol.name == "CA")]
            if len(ca1coor) and len(ca2coor):
                dist = np.sqrt(np.sum((ca1coor.squeeze() - ca2coor.squeeze()) ** 2))
                if dist < spatialgap:
                    todelete.append(i)
            i += 1
        todelete = np.array(todelete, dtype=int)
        # Join the non-real gaps into segments
        idxstartseg = np.delete(idxstartseg, todelete + 1)
        idxendseg = np.delete(idxendseg, todelete)

        mol.set("resid", residbackup)  # Restoring the original resids

    for i, (s, e) in enumerate(zip(idxstartseg, idxendseg)):
        if "chain" in fields:
            newchainid = available_chains[i % len(available_chains)]
            if _logger:
                logger.info(
                    f"Set chain {newchainid} between resid {mol.resid[s]} and {mol.resid[e]}."
                )
            mol.chain[s : e + 1] = newchainid
        if "segid" in fields:
            newsegid = basename + str(i)
            if _logger:
                logger.info(
                    f"Created segment {newsegid} between resid {mol.resid[s]} and {mol.resid[e]}."
                )
            mol.segid[s : e + 1] = newsegid

    return mol


def autoSegment2(
    mol,
    sel="(protein or resname ACE NME)",
    basename="P",
    fields=("segid",),
    residgaps=False,
    residgaptol=1,
    chaingaps=True,
    _logger=True,
):
    """Detects bonded segments in a selection and assigns incrementing segid to each segment

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule object
    sel : str
        Atom selection string on which to check for gaps.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    basename : str
        The basename for segment ids. For example if given 'P' it will name the segments 'P1', 'P2', ...
    fields : tuple of strings
        Field to fix. Can be "segid" (default) or any other Molecule field or combinations thereof.
    residgaps : bool
        Set to True to consider gaps in resids as structural gaps. Set to False to ignore resids
    residgaptol : int
        Above what resid difference is considered a gap. I.e. with residgaptol 1, 235-233 = 2 > 1 hence is a gap. We set
        default to 2 because in many PDBs single residues are missing in the proteins without any gaps.
    chaingaps : bool
        Set to True to consider changes in chains as structural gaps. Set to False to ignore chains

    Returns
    -------
    newmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        A new Molecule object with modified segids

    Example
    -------
    >>> newmol = autoSegment2(mol)
    """
    from scipy.sparse import csr_matrix
    from scipy.sparse.csgraph import connected_components

    if isinstance(fields, str):
        fields = (fields,)

    orig_sel = sel
    sel += " and backbone or (resname NME ACE and name N C O CH3)"  # Looking for bonds only over the backbone of the protein
    # Keep the original atom indexes to map from submol to mol
    idx = mol.atomselect(sel, indexes=True)
    # We filter out everything not on the backbone to calculate only those bonds
    submol = mol.copy()
    submol.filter(sel, _logger=False)
    bonds = submol._getBonds()  # Calculate both file and guessed bonds

    if residgaps:
        # Remove bonds between residues without continuous resids
        bondresiddiff = np.abs(submol.resid[bonds[:, 0]] - submol.resid[bonds[:, 1]])
        bonds = bonds[bondresiddiff <= residgaptol, :]
    else:
        # Warning about bonds bonding non-continuous resids
        bondresiddiff = np.abs(submol.resid[bonds[:, 0]] - submol.resid[bonds[:, 1]])
        if _logger and np.any(bondresiddiff > 1):
            for i in np.where(bondresiddiff > residgaptol)[0]:
                logger.warning(
                    "Bonds found between resid gaps: resid {} and {}".format(
                        submol.resid[bonds[i, 0]], submol.resid[bonds[i, 1]]
                    )
                )
    if chaingaps:
        # Remove bonds between residues without same chain
        bondsamechain = submol.chain[bonds[:, 0]] == submol.chain[bonds[:, 1]]
        bonds = bonds[bondsamechain, :]
    else:
        # Warning about bonds bonding different chains
        bondsamechain = submol.chain[bonds[:, 0]] == submol.chain[bonds[:, 1]]
        if _logger and np.any(bondsamechain == False):
            for i in np.where(bondsamechain == False)[0]:
                logger.warning(
                    "Bonds found between chain gaps: resid {}/{} and {}/{}".format(
                        submol.resid[bonds[i, 0]],
                        submol.chain[bonds[i, 0]],
                        submol.resid[bonds[i, 1]],
                        submol.chain[bonds[i, 1]],
                    )
                )

    # Calculate connected components using the bonds
    sparsemat = csr_matrix(
        (
            np.ones(bonds.shape[0] * 2),  # Values
            (
                np.hstack((bonds[:, 0], bonds[:, 1])),  # Rows
                np.hstack((bonds[:, 1], bonds[:, 0])),
            ),
        ),
        shape=[submol.numAtoms, submol.numAtoms],
    )  # Columns
    numcomp, compidx = connected_components(sparsemat, directed=False)

    # Letters to be used for chains, if free: 0123456789abcd...ABCD..., minus chain symbols already used
    sel_mask = mol.atomselect(orig_sel)
    used_chains = set(mol.chain[~sel_mask])
    available_chains = [x for x in chain_alphabet if x not in used_chains]
    used_segids = set([x[0] for x in mol.segid[~sel_mask] if x != ""])
    available_segids = [x for x in [basename] + segid_alphabet if x not in used_segids]
    basename = available_segids[0]

    mol = mol.copy()
    prevsegres = None
    for i in range(numcomp):  # For each connected component / segment
        segid = basename + str(i)
        backboneSegIdx = idx[compidx == i]  # The backbone atoms of the segment
        segres = mol.atomselect(
            "same residue as index {}".format(" ".join(map(str, backboneSegIdx)))
        )  # Get whole residues

        # Warning about separating segments with continuous resids
        if (
            _logger
            and i > 0
            and (np.min(mol.resid[segres]) - np.max(mol.resid[prevsegres])) == 1
        ):
            logger.warning(
                f"Separated segments {basename + str(i - 1)} and {segid}, despite continuous resids, due to lack of bonding."
            )

        # Add the new segment ID to all fields the user specified
        for f in fields:
            if f != "chain":
                mol.__dict__[f][segres] = segid  # Assign the segid to the correct atoms
            else:
                mol.__dict__[f][segres] = available_chains[i % len(available_chains)]
        if _logger:
            logger.info(
                f"Created segment {segid} between resid {mol.resid[segres].min()} and {mol.resid[segres].max()}."
            )
        prevsegres = segres  # Store old segment atom indexes for the warning about continuous resids

    return mol


import unittest


class _TestPreparation(unittest.TestCase):
    def test_autoSegment(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from os import path

        p = Molecule(path.join(home(dataDir="test-autosegment"), "4dkl.pdb"))
        p.filter("(chain B and protein) or water")
        p = autoSegment(p, "protein", "P")
        m = Molecule(path.join(home(dataDir="test-autosegment"), "membrane.pdb"))
        print(np.unique(m.get("segid")))

        mol = Molecule(path.join(home(dataDir="test-autosegment"), "1ITG_clean.pdb"))
        ref = Molecule(path.join(home(dataDir="test-autosegment"), "1ITG.pdb"))
        mol = autoSegment(mol, sel="protein")
        assert np.all(mol.segid == ref.segid)

        mol = Molecule(path.join(home(dataDir="test-autosegment"), "3PTB_clean.pdb"))
        ref = Molecule(path.join(home(dataDir="test-autosegment"), "3PTB.pdb"))
        mol = autoSegment(mol, sel="protein")
        assert np.all(mol.segid == ref.segid)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
import ctypes as ct
import numpy as np

from moleculekit.home import home


def wrap(coords, bonds, box, centersel=None):
    """
    Wrap the coords back into the unit cell. 
    Molecules will remain continuous, so they may escape the bounds of the prinary unit cell.
    """

    import platform
    libdir = home(libDir=True)

    if coords.dtype != np.float32:
        raise ValueError("coords is not float32")

    if coords.ndim == 2:
        coords = coords[:, :, None]

    if coords.shape[1] != 3:
        raise RuntimeError("coords needs to be natoms x 3 x nframes")

    nframes = coords.shape[2]
    natoms = coords.shape[0]
    nbonds = bonds.shape[0]

    if coords.strides[0] != 12 * nframes or coords.strides[1] != 4 * nframes:
        # It's a view -- need to make a copy to ensure contiguity of memory
        coords = np.array(coords, dtype=np.float32)
    if coords.strides[0] != 12 * nframes or coords.strides[1] != 4 * nframes:
        raise ValueError("coords is a view with unsupported strides")

    if np.size(bonds, 1) != 2:
        raise RuntimeError("'bonds' not nbonds x 2 in length")
    if np.size(box, 0) != 3 or np.size(box, 1) != nframes:
        raise RuntimeError("'box' not 3 x nframes in length")

    if platform.system() == "Windows":
        ct.cdll.LoadLibrary(os.path.join(libdir, "libgcc_s_seh-1.dll"))
        if os.path.exists(os.path.join(libdir, "psprolib.dll")):
            ct.cdll.LoadLibrary(os.path.join(libdir, "psprolib.dll"))

    lib = ct.cdll.LoadLibrary(os.path.join(libdir, "libvmdparser.so"))

    c_nbonds = ct.c_int(nbonds)
    c_natoms = ct.c_int(natoms)
    c_nframes = ct.c_int(nframes)

    if centersel is None:
        centersel = np.array([-1], dtype=np.int32)
    centersel = np.append(centersel, np.array([-1], dtype=np.int32))
    c_centersel = centersel.ctypes.data_as(ct.POINTER(ct.c_int))

    c_coords = coords.ctypes.data_as(ct.POINTER(ct.c_float))
    c_bonds = bonds.flatten().astype(np.int32).copy().ctypes.data_as(ct.POINTER(ct.c_int))
    c_box = box.T.flatten().astype(np.float64).copy().ctypes.data_as(ct.POINTER(ct.c_double))

    lib.wrap(
        c_bonds,
        c_coords,
        c_box, 
        c_nbonds, 
        c_natoms, 
        c_nframes, 
        c_centersel
    )

    return coords


from unittest import TestCase
class _TestWrap(TestCase):
    def test_wrap(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home

        datadir = os.path.join(home(dataDir='molecule-readers'), 'multi-traj')
        mol = Molecule(os.path.join(datadir, 'structure.pdb'))
        mol.read(os.path.join(datadir, 'data', 'e1s1_1', 'output.xtc'))
        mol.wrap()

        boxsize = mol.coords.max(axis=0) - mol.coords.min(axis=0)

        assert np.abs(mol.box - boxsize).max() < 2


if __name__ == "__main__":
    import unittest

    unittest.main(verbosity=2)


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
try:
    import networkx as nx
except ImportError as e:
    print('Could not import networkx which is necessary for graph alignment. You can install it with `conda install networkx`.')

import numpy as np
from unittest import TestCase
import os


def createProductGraph(G, H, tolerance, fields):
    # Calculate product graph by creating a node for each feature-matching pair of points
    newnodes = []
    for gn in G.nodes():
        for hn in H.nodes():
            matching = True
            for f in fields:
                if G.nodes[gn][f] != H.nodes[hn][f]:
                    matching = False
                    break
            if matching:
                newnodes.append((gn, hn))

    Gprod = nx.Graph()
    Gprod.add_nodes_from(newnodes)

    # Add edges when distances in both graphs agree within tolerance error between the two given nodes
    for np1 in range(len(newnodes)):
        for np2 in range(np1 + 1, len(newnodes)):
            pair1 = newnodes[np1]
            pair2 = newnodes[np2]
            if not G.has_edge(pair1[0], pair2[0]) or not H.has_edge(pair1[1], pair2[1]):
                continue
            dist1 = G.edges[pair1[0], pair2[0]]['distance']
            dist2 = H.edges[pair1[1], pair2[1]]['distance']
            if abs(dist1 - dist2) < tolerance:
                Gprod.add_edge(newnodes[np1], newnodes[np2])
    return Gprod


def compareGraphs(G, H, fields=('element',), tolerance=0.5, returnmatching=False):
    # Comparison algorithm based on:
    # "Chemoisosterism in the Proteome", X. Jalencas, J. Mestres, JCIM 2013
    # http://pubs.acs.org/doi/full/10.1021/ci3002974
    if G == H:
        if returnmatching:
            return 1, len(G), [(x, x) for x in G.nodes()]
        else:
            return 1

    if len(G.edges()) == 0 or len(H.edges()) == 0:
        if returnmatching:
            return 0, 0, []
        else:
            return 0

    Gprod = createProductGraph(G, H, tolerance, fields)

    # Calculate the maximal cliques and return the length of the largest one
    maxcliques = np.array(list(nx.find_cliques(Gprod)))
    cllen = np.array([len(x) for x in maxcliques])
    score = (cllen.max() / max(len(G.nodes()), len(H.nodes())))

    if returnmatching:
        return score, cllen.max(), maxcliques[cllen.argmax()]
    else:
        return score


def makeMolGraph(mol, sel, fields):
    from scipy.spatial.distance import pdist, squareform

    if sel != 'all':
        sel = mol.atomselect(sel, indexes=True)
    else:
        sel = np.arange(mol.numAtoms)

    g = nx.Graph()
    for i in sel:
        props = {f: mol.__dict__[f][i] for f in fields}
        g.add_node(i, **props)

    distances = squareform(pdist(mol.coords[sel, :, mol.frame]))
    nodes = list(g.nodes())
    for i in range(len(g)):
        for j in range(i+1, len(g)):
            g.add_edge(nodes[i], nodes[j], distance=distances[i, j])

    return g


def maximalSubstructureAlignment(mol1, mol2, sel1='all', sel2='all', fields=('element',), tolerance=0.5, visualize=False):
    """ Aligns two molecules on the largest common substructure

    Parameters
    ----------
    mol1 : :class:`Molecule`
        The reference molecule on which to align
    mol2 : :class:`Molecule`
        The second molecule which will be rotated and translated to align on mol1
    sel1 : str
        Atom selection string of the atoms of `mol1` to align.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    sel2 : str
        Atom selection string of the atoms of `mol2` to align.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    fields : tuple
        A tuple of the fields that are used to match atoms
    tolerance : float
        How different can distances be between to atom pairs for them to match in the product graph
    visualize : bool
        If set to True it will visualize the alignment

    Returns
    -------
    newmol : :class:`Molecule`
        A copy of mol2 aligned on mol1
    """
    mol2 = mol2.copy()
    g1 = makeMolGraph(mol1, sel1, fields)
    g2 = makeMolGraph(mol2, sel2, fields)

    _, _, matching = compareGraphs(g1, g2, fields=fields, tolerance=tolerance, returnmatching=True)

    matchnodes1 = np.array([x[0] for x in matching])
    matchnodes2 = np.array([x[1] for x in matching])

    mol2.align(sel=matchnodes2, refmol=mol1, refsel=matchnodes1)

    if visualize:
        mol1.view(sel='index {}'.format(' '.join(map(str, matchnodes1))), style='CPK', hold=True)
        mol1.view(sel='all', style='Lines')

        mol2.view(sel='index {}'.format(' '.join(map(str, matchnodes2))), style='CPK', hold=True)
        mol2.view(sel='all', style='Lines')

    return mol2


class _TestGraphAlignment(TestCase):
    def test_maximalSubstructureAlignment(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule

        path = home(dataDir='test-molecule-graphalignment')
        ref_lig = Molecule(os.path.join(path, 'ref_lig.pdb'))
        lig2align = Molecule(os.path.join(path, 'lig2align.pdb'))
        lig_aligned = maximalSubstructureAlignment(ref_lig, lig2align)
        lig_reference = Molecule(os.path.join(path, 'lig_aligned.pdb'))

        self.assertTrue(np.allclose(lig_aligned.coords, lig_reference.coords, rtol=1e-4),
                        'maximalSubstructureAlignment produced different coords')

if __name__ == "__main__":
    import unittest

    unittest.main(verbosity=2)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.projection import Projection
import numpy as np
import logging

logger = logging.getLogger(__name__)


class MetricCoordinate(Projection):
    """Creates a MetricCoordinate object that calculates the atom coordinates from a set of trajectories.

    Parameters
    ----------
    atomsel : str
        Atom selection string for the atoms whose coordinates we want to calculate.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    refmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The reference Molecule to which we will align.
    trajalnsel : str, optional
        Atom selection string for the trajectories from which to align to the reference structure.
        If it's None and a `refmol` is passed it will default to 'protein and name CA'.
    refalnsel : str, optional
        Atom selection string for `refmol` from which to align to the reference structure. If None, it defaults to the
        same as `trajalnsel`.
    centersel : str, optional
        Atom selection string around which to wrap the simulation.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    pbc : bool
        Enable or disable coordinate wrapping based on periodic boundary conditions.

    Returns
    -------
    metr : MetricCoordinate object
    """

    def __init__(
        self,
        atomsel,
        refmol=None,
        trajalnsel=None,
        refalnsel=None,
        centersel="protein",
        pbc=True,
    ):
        super().__init__()

        if atomsel is None:
            raise ValueError("Atom selection cannot be None")

        self._trajalnsel = trajalnsel
        self._centersel = centersel
        self._atomsel = atomsel
        self._pbc = pbc
        self._refmol = refmol

        if self._refmol is not None:
            if self._trajalnsel is None:
                self._trajalnsel = "protein and name CA"
            self._refalnsel = refalnsel if refalnsel is not None else self._trajalnsel
            self._refmol = refmol.copy()
            self._cache["refalnsel"] = self._refmol.atomselect(self._refalnsel)

    def _calculateMolProp(self, mol, props="all"):
        props = ("trajalnsel", "atomsel", "centersel") if props == "all" else props
        res = {}
        if "trajalnsel" in props:
            res["trajalnsel"] = None
            if self._trajalnsel is not None:
                res["trajalnsel"] = mol.atomselect(self._trajalnsel)
                if np.sum(res["trajalnsel"]) == 0:
                    raise RuntimeError("Alignment selection resulted in 0 atoms.")
        if "atomsel" in props:
            res["atomsel"] = mol.atomselect(self._atomsel)
            if np.sum(res["atomsel"]) == 0:
                raise RuntimeError("Atom selection resulted in 0 atoms.")
        if "centersel" in props and self._pbc:
            res["centersel"] = mol.atomselect(self._centersel)
            if np.sum(res["centersel"]) == 0:
                raise RuntimeError("Centering selection resulted in 0 atoms.")
        return res

    def project(self, mol):
        """Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)

        mol = mol.copy()
        if self._pbc:
            mol.wrap(getMolProp("centersel"))

        trajalnsel = getMolProp("trajalnsel")

        if trajalnsel is not None:
            if self._refmol is None:
                mol.align(trajalnsel)
            else:
                mol.align(
                    trajalnsel, refmol=self._refmol, refsel=getMolProp("refalnsel")
                )

        coords = np.transpose(
            mol.coords[getMolProp("atomsel")], axes=(2, 1, 0)
        ).reshape(mol.numFrames, -1)
        return coords

    def getMapping(self, mol):
        """Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        getMolProp = lambda prop: self._getMolProp(mol, prop)

        atomidx = np.where(getMolProp("atomsel"))[0]
        from pandas import DataFrame

        types = []
        indexes = []
        description = []
        for xyz in ("X", "Y", "Z"):
            for i in atomidx:
                types += ["coordinate"]
                indexes += [i]
                description += [
                    "{} coordinate of {} {} {}".format(
                        xyz, mol.resname[i], mol.resid[i], mol.name[i]
                    )
                ]
        return DataFrame(
            {"type": types, "atomIndexes": indexes, "description": description}
        )


import unittest


class _TestMetricCoordinate(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from os import path

        mol = Molecule(
            path.join(home(dataDir="test-projections"), "trajectory", "filtered.pdb")
        )
        mol.read(path.join(home(dataDir="test-projections"), "trajectory", "traj.xtc"))

        self.mol = mol

    def test_project(self):
        ref = self.mol.copy()
        ref.coords = np.atleast_3d(ref.coords[:, :, 0])
        metr = MetricCoordinate("protein and name CA")
        data = metr.project(self.mol)

        lastcoors = np.array(
            [
                -24.77000237,
                -27.76000023,
                -30.44000244,
                -33.65000153,
                -33.40999985,
                -36.32000351,
                -36.02000427,
                -36.38000107,
                -39.61000061,
                -41.01000214,
                -43.80000305,
                -45.56000137,
                -45.36000061,
                -47.13000488,
                -49.54000473,
                -50.6000061,
                -50.11000061,
                -52.15999985,
                -55.1400032,
                -55.73000336,
            ],
            dtype=np.float32,
        )
        assert np.all(
            np.abs(data[-1, -20:] - lastcoors) < 0.001
        ), "Coordinates calculation is broken"

    def test_project_align(self):
        ref = self.mol.copy()
        ref.coords = np.atleast_3d(ref.coords[:, :, 0])
        metr = MetricCoordinate("protein and name CA", ref)
        data = metr.project(self.mol)

        lastcoors = np.array(
            [
                6.79283285,
                5.55226946,
                4.49387407,
                2.94484425,
                5.36937141,
                3.18590879,
                5.75874281,
                5.48864174,
                1.69625032,
                1.58790839,
                0.57877392,
                -2.66498065,
                -3.70919156,
                -3.33702421,
                -5.38465405,
                -8.43286991,
                -8.15859032,
                -7.85062265,
                -10.92551327,
                -13.70733166,
            ],
            dtype=np.float32,
        )
        assert np.all(
            np.abs(data[-1, -20:] - lastcoors) < 0.001
        ), "Coordinates calculation is broken"


if __name__ == "__main__":
    unittest.main(verbosity=2)


import networkx as nx
import numpy as np
from moleculekit.tools.moleculechecks import isProteinProtonated
import logging

logger = logging.getLogger(__name__)


def get_donors_acceptors(mol, exclude_water=True, exclude_backbone=False):
    if not isProteinProtonated(mol):
        raise RuntimeError(
            "The protein seems to not be protonated. You must provide a protonated system for H-bonds to be detected."
        )

    donor_elements = ("N", "O")
    acceptor_elements = ("N", "O")
    potential_donors = np.in1d(mol.element, donor_elements)
    potential_acceptors = np.in1d(mol.element, acceptor_elements)

    if not np.any(potential_donors) and not np.any(potential_acceptors):
        logger.warning("Could not find any [N, O] elements in the molecule")
        return [], []

    backbone = mol.atomselect("protein and backbone")
    # Amide nitrogens are only donors
    potential_acceptors[(mol.element == "N") & backbone] = False

    if exclude_water:
        waters = mol.atomselect("water")
        potential_donors = ~waters & potential_donors
        potential_acceptors = ~waters & potential_acceptors

    if exclude_backbone:
        potential_donors = ~backbone & potential_donors
        potential_acceptors = ~backbone & potential_acceptors

    if not np.any(potential_donors):
        logger.warning("Did not find any potential donors")
        return [], []

    if not np.any(potential_acceptors):
        logger.warning("Did not find any potential acceptors")
        return [], []

    hydrogens = tuple(np.where(mol.element == "H")[0])
    donor_pairs = []
    potential_donors = np.where(potential_donors)[0]

    mask1 = np.isin(mol.bonds[:, 0], potential_donors) & np.isin(
        mol.bonds[:, 1], hydrogens
    )
    mask2 = np.isin(mol.bonds[:, 1], potential_donors) & np.isin(
        mol.bonds[:, 0], hydrogens
    )
    donor_pairs1 = mol.bonds[mask1, :]
    donor_pairs2 = mol.bonds[
        mask2, ::-1
    ]  # Invert the order here to have the hyd second
    donor_pairs = np.vstack((donor_pairs1, donor_pairs2))

    acceptors = np.where(potential_acceptors)[0]

    if not len(donor_pairs) or not len(acceptors):
        return [], []
    return donor_pairs.astype(np.uint32), acceptors.astype(np.uint32)


def view_hbonds(mol, hbonds):
    from moleculekit.vmdgraphics import VMDCylinder

    viewname = "mol"
    if mol.viewname is not None:
        viewname = mol.viewname

    for f in range(mol.numFrames):
        mol.viewname = f"{viewname}_frame_{f}"
        mol.view()
        for i in range(hbonds[f].shape[0]):
            end = hbonds[f][i, 2]
            start = hbonds[f][i, 1]
            if start == -1:
                start = hbonds[f][i, 0]

            VMDCylinder(
                mol.coords[start, :, f],
                mol.coords[end, :, f],
                radius=0.1,
            )
    mol.viewname = viewname


def get_protein_rings(mol):
    import networkx as nx

    _aromatics = ["PHE", "HIS", "HID", "HIE", "HIP", "TYR", "TRP"]
    _excluded_atoms = ["N", "CA", "C", "O"]  # , "CB", "OH"]

    arom_res_atoms = np.isin(mol.resname, _aromatics) & ~np.isin(
        mol.name, _excluded_atoms
    )
    mol2 = mol.copy()
    mol2.filter(arom_res_atoms, _logger=False)
    bonds = mol2._guessBonds()

    graph = nx.Graph()
    graph.add_edges_from(bonds)
    cycles = nx.cycle_basis(graph)

    original_idx = np.where(arom_res_atoms)[0]
    cycles = [tuple(original_idx[cc]) for cc in cycles]

    return cycles


def get_nucleic_rings(mol):
    import networkx as nx

    nucleic_sel = mol.atomselect("nucleic and not backbone")
    original_idx = np.where(nucleic_sel)[0]

    mol2 = mol.copy()
    mol2.filter(nucleic_sel, _logger=False)
    bonds = mol2._guessBonds()

    graph = nx.Graph()
    graph.add_edges_from(bonds)
    cycles = nx.cycle_basis(graph)

    cycles = [tuple(original_idx[cc]) for cc in cycles]

    return cycles


def get_protein_aryl_halides(mol):
    # TODO: Need to support non-standard residues
    pass


metals = [
    "Fe",
    "Cu",
    "Ni",
    "Mo",
    "Rh",
    "Re",
    "Mn",
    "Mg",
    "Ca",
    "Na",
    "K",
    "Cs",
    "Zn",
    "Se",
]


def get_metal_charged(mol):
    return list(np.where(np.isin(mol.element, metals))[0]), []


def get_protein_charged(mol):
    lys_n = (mol.resname == "LYS") & (mol.name == "NZ")
    arg_c = (mol.resname == "ARG") & (mol.name == "CZ")
    hip_c = (mol.resname == "HIP") & (mol.name == "CE1")
    pos = lys_n | arg_c | hip_c

    asp_c = (mol.resname == "ASP") & (mol.name == "CG")
    glu_c = (mol.resname == "GLU") & (mol.name == "CD")
    neg = asp_c | glu_c

    return list(np.where(pos)[0]), list(np.where(neg)[0])


def get_nucleic_charged(mol):
    nuc = mol.atomselect("nucleic and backbone and name OP2")
    return [], list(np.where(nuc)[0])


def get_ligand_rings(sm, start_idx=0):
    ligandRings = sm._mol.GetRingInfo().AtomRings()
    ligandAtomAromaticRings = []
    for ring in ligandRings:
        aromatics = sum([sm._mol.GetAtomWithIdx(idx).GetIsAromatic() for idx in ring])
        if aromatics != len(ring):
            continue
        ligandAtomAromaticRings.append(tuple([r + start_idx for r in ring]))
    return ligandAtomAromaticRings


def get_ligand_charged(sm, start_idx=0):
    pos = []
    neg = []
    for i in range(sm.numAtoms):
        fc = sm._mol.GetAtomWithIdx(i).GetFormalCharge()
        if fc > 0:
            pos.append(i + start_idx)
        elif fc < 0:
            neg.append(i + start_idx)
    return pos, neg


def get_ligand_aryl_halides(sm, start_idx=0):
    import networkx as nx

    halogens = ["Cl", "Br", "I"]
    rings = get_ligand_rings(sm)
    if len(rings) == 0:
        return []

    ring_atoms = np.hstack(rings)
    halogens = np.where(np.isin(sm._element, halogens))[0]

    graph = nx.Graph()
    graph.add_edges_from(sm._bonds)

    halides = []
    for hi in halogens:
        neighs = list(graph.neighbors(hi))
        if len(neighs) != 1:
            continue
        if neighs[0] in ring_atoms:
            halides.append([hi + start_idx, neighs[0] + start_idx])

    return halides


def hbonds_calculate(
    mol,
    donors,
    acceptors,
    sel1="all",
    sel2=None,
    dist_threshold=2.5,
    angle_threshold=120,
    ignore_hs=False,
):
    from moleculekit.interactions import hbonds

    if len(donors) == 0 or len(acceptors) == 0:
        return [[] for _ in range(mol.numFrames)]

    sel1 = mol.atomselect(sel1).astype(np.uint32).copy()
    if sel2 is None:
        sel2 = sel1.copy()
        intra = True
    else:
        sel2 = mol.atomselect(sel2).astype(np.uint32).copy()
        intra = False

    if len(sel1) != mol.numAtoms or len(sel2) != mol.numAtoms:
        raise RuntimeError(
            "Selections must be boolean of size equal to number of atoms in the molecule"
        )

    # Filter donors and acceptors list if they are not in the selections to reduce calculations
    sel_idx = np.where(sel1 | sel2)[0]
    donors = donors[np.all(np.isin(donors, sel_idx), axis=1)]
    acceptors = acceptors[np.isin(acceptors, sel_idx)]

    hb = hbonds.calculate(
        donors,
        acceptors,
        mol.coords,
        mol.box,
        sel1,
        sel2,
        dist_threshold=dist_threshold,
        angle_threshold=angle_threshold,
        intra=intra,
        ignore_hs=ignore_hs,
    )

    hbond_list = []
    for f in range(mol.numFrames):
        hbond_list.append([hb[f][i : i + 3] for i in range(0, len(hb[f]), 3)])
    return hbond_list


def waterbridge_calculate(
    mol,
    donors,
    acceptors,
    sel1,
    sel2,
    order=1,
    dist_threshold=2.5,
    angle_threshold=120,
    ignore_hs=False,
):
    import networkx as nx

    if len(donors) == 0 or len(acceptors) == 0:
        return [[] for _ in range(mol.numFrames)]

    sel1_b = mol.atomselect(sel1)
    sel2_b = mol.atomselect(sel2)
    water_b = mol.atomselect("water")

    args = {
        "mol": mol,
        "donors": donors,
        "acceptors": acceptors,
        "dist_threshold": dist_threshold,
        "angle_threshold": angle_threshold,
        "ignore_hs": ignore_hs,
    }

    water_goal = water_b | sel2_b
    water_goal_idx = np.where(water_goal)[0]
    water_idx = np.where(water_b)[0]

    # Add one because an order 1 water bridge requires two iterations to find the target
    order += 1
    edges = []
    for f in range(mol.numFrames):
        edges.append([])

    sel1_b_curr = sel1_b.copy()
    for _ in range(order):
        curr_shell = hbonds_calculate(sel1=sel1_b_curr, sel2=water_goal, **args)
        for f in range(mol.numFrames):
            # Only keep interactions which have at least one water
            curr_shell[f] = np.array(curr_shell[f])
            has_water = np.any(np.isin(curr_shell[f], water_idx), axis=1)
            curr_shell[f] = curr_shell[f][has_water, :]
            # Append the valid edges for this frame
            if ignore_hs:
                edges[f].append(curr_shell[f][:, [0, 2]])
            else:
                edges[f].append(curr_shell[f][:, :2])
                edges[f].append(curr_shell[f][:, 1:])

        # Find which water atoms interacted with sel1_b_curr to use them for the next shell
        shell = np.vstack(curr_shell)[:, [0, 2]]
        sel1_b_curr = np.zeros(mol.numAtoms, dtype=bool)
        interacted = np.unique(shell[np.isin(shell, water_goal_idx)])
        if len(interacted) == 0:
            break
        sel1_b_curr[interacted] = True

    # Create networks and check for shortest paths between source and target
    sel1_idx = np.where(sel1_b)[0]
    sel2_idx = np.where(sel2_b)[0]

    water_bridges = []
    for f in range(mol.numFrames):
        water_bridges.append([])

    for f in range(mol.numFrames):
        ee = np.vstack(edges[f])
        starts = np.unique(ee[np.isin(ee, sel1_idx)])
        ends = np.unique(ee[np.isin(ee, sel2_idx)])

        # If they never connected skip
        if not (np.any(starts) and np.any(ends)):
            continue

        network = nx.Graph()
        network.add_edges_from(ee)
        # For all start and end indexes (sel1/sel2 can have multiple donors/acceptors)
        for st in starts:
            for en in ends:
                for pp in nx.all_simple_paths(network, source=st, target=en):
                    # Exclude direct source-target hbonds. Probably not needed due to the previous water check
                    if len(pp) < 3:
                        continue
                    # Exclude paths which are not pure water bridges
                    if not np.all(np.isin(pp[1:-1], water_idx)):
                        continue
                    water_bridges[f].append(pp)

    return water_bridges


def pipi_calculate(
    mol,
    rings1,
    rings2,
    dist_threshold1=4.4,
    angle_threshold1_max=30,
    dist_threshold2=5.5,
    angle_threshold2_min=60,
    return_rings=False,
):
    from moleculekit.interactions import pipi

    if (
        angle_threshold1_max < 0
        or angle_threshold1_max > 90
        or angle_threshold2_min < 0
        or angle_threshold2_min > 90
    ):
        raise RuntimeError("Values for angles should be [0, 90] degrees")

    if len(rings1) == 0 or len(rings2) == 0:
        return [[] for _ in range(mol.numFrames)], [[] for _ in range(mol.numFrames)]

    ring_atoms = np.hstack((np.hstack(rings1), np.hstack(rings2)))
    ring_starts1 = np.insert(np.cumsum([len(rr) for rr in rings1]), 0, 0)
    ring_starts2 = np.insert(np.cumsum([len(rr) for rr in rings2]), 0, 0)
    ring_starts2 += ring_starts1.max()

    pp, da = pipi.calculate(
        ring_atoms.astype(np.uint32),
        ring_starts1.astype(np.uint32),
        ring_starts2.astype(np.uint32),
        mol.coords,
        mol.box,
        dist_threshold1=dist_threshold1,
        angle_threshold1_max=angle_threshold1_max,
        dist_threshold2=dist_threshold2,
        angle_threshold2_min=angle_threshold2_min,
    )

    pp_list = []
    dist_ang_list = []
    for f in range(mol.numFrames):
        reshaped = [pp[f][i : i + 2] for i in range(0, len(pp[f]), 2)]
        if return_rings:
            pp_list.append([[rings1[pp[0]], rings2[pp[1]]] for pp in reshaped])
        else:
            pp_list.append(reshaped)
        dist_ang_list.append([da[f][i : i + 2] for i in range(0, len(da[f]), 2)])
    return pp_list, dist_ang_list


def saltbridge_calculate(mol, pos, neg, sel1="all", sel2=None, threshold=4):
    from moleculekit.projections.util import pp_calcDistances

    if len(pos) == 0 or len(neg) == 0:
        return [[] for _ in range(mol.numFrames)]

    sel1 = mol.atomselect(sel1).astype(np.uint32).copy()
    if sel2 is None:
        sel2 = sel1.copy()
    else:
        sel2 = mol.atomselect(sel2).astype(np.uint32).copy()

    charged = np.zeros(sel1.shape, dtype=bool)
    charged[pos] = True
    charged[neg] = True

    sel1 = sel1 & charged
    sel2 = sel2 & charged

    if np.sum(sel1) == 0 or np.sum(sel2) == 0:
        return [[] for _ in range(mol.numFrames)]

    periodic = "selections" if not np.all(mol.box == 0) else None
    dists = pp_calcDistances(mol, sel1, sel2, periodic, metric="distances")

    sel1 = np.where(sel1)[0]
    sel2 = np.where(sel2)[0]

    salt_bridge_list = []
    for f in range(mol.numFrames):
        idx = np.where(dists[f] < threshold)[0]
        sel1_sub = sel1[(idx % len(sel1)).astype(int)]
        sel2_sub = sel2[(idx / len(sel1)).astype(int)]
        inter = np.vstack((sel1_sub, sel2_sub)).T
        # Should only have one positive per row (the other is necessarily negative)
        opposite_charges = np.sum(np.isin(inter, pos), axis=1) == 1
        salt_bridge_list.append(inter[opposite_charges])

    return salt_bridge_list


def cationpi_calculate(
    mol,
    rings,
    cations,
    dist_threshold=5,
    angle_threshold_min=60,
    return_rings=False,
):
    from moleculekit.interactions import cationpi

    if angle_threshold_min < 0 or angle_threshold_min > 90:
        raise RuntimeError("Values for angles should be [0, 90] degrees")

    if len(rings) == 0 or len(cations) == 0:
        return [[] for _ in range(mol.numFrames)], [[] for _ in range(mol.numFrames)]

    ring_atoms = np.hstack(rings)
    ring_starts = np.insert(np.cumsum([len(rr) for rr in rings]), 0, 0)

    pp, da = cationpi.calculate(
        ring_atoms.astype(np.uint32),
        ring_starts.astype(np.uint32),
        np.array(cations, dtype=np.uint32),
        mol.coords,
        mol.box,
        dist_threshold=dist_threshold,
        angle_threshold_min=angle_threshold_min,
    )

    index_list = []
    dist_ang_list = []
    for f in range(mol.numFrames):
        reshaped = [pp[f][i : i + 2] for i in range(0, len(pp[f]), 2)]
        if return_rings:
            index_list.append([[rings[pp[0]], pp[1]] for pp in reshaped])
        else:
            index_list.append(reshaped)
        dist_ang_list.append([da[f][i : i + 2] for i in range(0, len(da[f]), 2)])
    return index_list, dist_ang_list


def sigmahole_calculate(
    mol,
    rings,
    halides,
    dist_threshold=4.5,
    angle_threshold_min=60,
    return_rings=False,
):
    from moleculekit.interactions import sigmahole

    if angle_threshold_min < 0 or angle_threshold_min > 90:
        raise RuntimeError("Values for angles should be [0, 90] degrees")

    if len(rings) == 0 or len(halides) == 0:
        return [[] for _ in range(mol.numFrames)], [[] for _ in range(mol.numFrames)]

    ring_atoms = np.hstack(rings)
    ring_starts = np.insert(np.cumsum([len(rr) for rr in rings]), 0, 0)

    indexes, da = sigmahole.calculate(
        ring_atoms.astype(np.uint32),
        ring_starts.astype(np.uint32),
        np.array(halides, dtype=np.uint32),
        mol.coords,
        mol.box,
        dist_threshold=dist_threshold,
        angle_threshold_min=angle_threshold_min,
    )

    index_list = []
    dist_ang_list = []
    for f in range(mol.numFrames):
        reshaped = [indexes[f][i : i + 2] for i in range(0, len(indexes[f]), 2)]
        if return_rings:
            index_list.append([[rings[pp[0]], pp[1]] for pp in reshaped])
        else:
            index_list.append(reshaped)
        dist_ang_list.append([da[f][i : i + 2] for i in range(0, len(da[f]), 2)])
    return index_list, dist_ang_list


import unittest


class _TestInteractions(unittest.TestCase):
    def test_hbonds(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from moleculekit.smallmol.smallmol import SmallMol
        import os

        prot = os.path.join(home(dataDir="test-interactions"), "3PTB_prepared.pdb")
        lig = os.path.join(home(dataDir="test-interactions"), "3PTB_BEN.sdf")

        mol = Molecule(prot)
        lig = SmallMol(lig).toMolecule()
        mol.append(lig)

        # TODO: Add check if bonds exist in molecule! Add it to moleculechecks
        mol.bonds = mol._guessBonds()
        mol.coords = np.tile(mol.coords, (1, 1, 2)).copy()  # Fake second frame
        donors, acceptors = get_donors_acceptors(
            mol, exclude_water=True, exclude_backbone=False
        )

        hb = hbonds_calculate(mol, donors, acceptors, "protein", "resname BEN")
        assert len(hb) == 2
        ref = np.array(
            [
                [3414, 3421, 2471],
                [3414, 3422, 2789],
                [3415, 3423, 2472],
                [3415, 3424, 2482],
            ]
        )
        assert np.array_equal(hb[0], ref) and np.array_equal(hb[1], ref), print(hb, ref)

        hb = hbonds_calculate(mol, donors, acceptors, "all")
        assert len(hb) == 2
        assert np.array(hb[0]).shape == (178, 3), np.array(hb[0]).shape

    def test_pipi(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from moleculekit.smallmol.smallmol import SmallMol
        import os

        mol = Molecule(os.path.join(home(dataDir="test-interactions"), "5L87.pdb"))
        lig = SmallMol(os.path.join(home(dataDir="test-interactions"), "5L87_6RD.sdf"))

        lig_idx = np.where(mol.resname == "6RD")[0][0]

        prot_rings = get_protein_rings(mol)
        lig_rings = get_ligand_rings(lig, start_idx=lig_idx)

        pipis, distang = pipi_calculate(mol, prot_rings, lig_rings)

        assert len(pipis) == 1

        ref_rings = np.array([[0, 2], [0, 4], [2, 4]])
        assert np.array_equal(pipis[0], ref_rings)

        ref_distang = np.array(
            [
                [5.33927107, 82.32684326],
                [5.23078251, 85.32985687],
                [5.16490269, 81.33213806],
            ]
        )
        assert np.allclose(distang[0], ref_distang)

        mol = Molecule(os.path.join(home(dataDir="test-interactions"), "6dn1.pdb"))
        lig = SmallMol(
            os.path.join(home(dataDir="test-interactions"), "6dn1_ligand-RDK.sdf")
        )

        lig_idx = np.where(mol.resname == "MOL")[0][0]

        prot_rings = get_nucleic_rings(mol)
        lig_rings = get_ligand_rings(lig, start_idx=lig_idx)

        pipis, distang = pipi_calculate(mol, prot_rings, lig_rings)

        assert len(pipis) == 1

        ref_rings = np.array([[72, 2], [73, 1], [74, 1]])
        assert np.array_equal(pipis[0], ref_rings)

        ref_distang = np.array(
            [
                [4.1462903, 22.34738922],
                [3.52108073, 11.22733593],
                [4.05929375, 4.60079527],
            ]
        )
        assert np.allclose(distang[0], ref_distang)

    def test_salt_bridge(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from moleculekit.smallmol.smallmol import SmallMol
        import os

        # 3ptb 5tvn 5l87

        prot = os.path.join(home(dataDir="test-interactions"), "3PTB_prepared.pdb")
        lig = os.path.join(home(dataDir="test-interactions"), "3PTB_BEN.sdf")

        mol = Molecule(prot)
        lig = SmallMol(lig)
        mol.append(lig.toMolecule())
        mol.coords = np.tile(mol.coords, (1, 1, 2)).copy()  # Fake second frame

        lig_idx = np.where(mol.resname == "BEN")[0][0]

        prot_pos, prot_neg = get_protein_charged(mol)
        lig_pos, lig_neg = get_ligand_charged(lig, start_idx=lig_idx)

        bridges = saltbridge_calculate(
            mol, prot_pos + lig_pos, prot_neg + lig_neg, "protein", "resname BEN"
        )

        assert len(bridges) == 2

        ref_bridge = np.array([[2470, 3414]])
        assert np.array_equal(bridges[0], ref_bridge)

    def test_cationpi_protein(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from moleculekit.smallmol.smallmol import SmallMol
        import os

        mol = Molecule(os.path.join(home(dataDir="test-interactions"), "1LPI.pdb"))
        prot_rings = get_protein_rings(mol)
        prot_pos, _ = get_protein_charged(mol)
        metal_pos, _ = get_metal_charged(mol)

        catpi, distang = cationpi_calculate(mol, prot_rings, prot_pos + metal_pos)

        ref_atms = np.array([[0, 8], [17, 1001], [18, 1001]])
        assert np.array_equal(ref_atms, catpi[0]), print(ref_atms, catpi[0])
        ref_distang = np.array(
            [
                [4.10110903, 63.69768524],
                [4.70270395, 60.51351929],
                [4.12248421, 82.81176758],
            ]
        )
        assert np.allclose(ref_distang, distang), distang

    def test_cationpi_protein_ligand(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from moleculekit.smallmol.smallmol import SmallMol
        import os

        lig = SmallMol(os.path.join(home(dataDir="test-interactions"), "2BOK_784.sdf"))
        mol = Molecule(
            os.path.join(home(dataDir="test-interactions"), "2BOK_prepared.pdb")
        )
        mol.append(lig.toMolecule())

        lig_idx = np.where(mol.resname == "784")[0][0]

        prot_rings = get_protein_rings(mol)
        lig_rings = get_ligand_rings(lig, start_idx=lig_idx)

        prot_pos, _ = get_protein_charged(mol)
        lig_pos, _ = get_ligand_charged(lig, start_idx=lig_idx)

        catpi, distang = cationpi_calculate(
            mol, prot_rings + lig_rings, prot_pos + lig_pos
        )

        ref_atms = np.array([[11, 3494]])
        assert np.array_equal(ref_atms, catpi[0]), print(ref_atms, catpi[0])
        ref_distang = np.array([[4.74848127, 74.07044983]])
        assert np.allclose(ref_distang, distang), distang

    def test_sigma_holes(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from moleculekit.smallmol.smallmol import SmallMol
        import os

        lig = SmallMol(os.path.join(home(dataDir="test-interactions"), "2P95_ME5.sdf"))
        mol = Molecule(
            os.path.join(home(dataDir="test-interactions"), "2P95_prepared.pdb")
        )
        mol.append(lig.toMolecule())

        lig_idx = np.where(mol.resname == "ME5")[0][0]
        lig_halides = get_ligand_aryl_halides(lig, start_idx=lig_idx)

        prot_rings = get_protein_rings(mol)

        sh, distang = sigmahole_calculate(mol, prot_rings, lig_halides)

        ref_atms = np.array([[29, 3702]])
        assert np.array_equal(ref_atms, sh[0]), print(ref_atms, sh[0])
        ref_distang = np.array([[4.26179695, 66.55052185]])
        assert np.allclose(ref_distang, distang)

    def test_water_bridge(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from moleculekit.smallmol.smallmol import SmallMol
        import os

        mol = Molecule(
            os.path.join(home(dataDir="test-interactions"), "5gw6_receptor_H_wet.pdb")
        )
        mol.bonds = mol._guessBonds()

        lig = SmallMol(
            os.path.join(home(dataDir="test-interactions"), "5gw6_ligand-RDK.sdf")
        )
        mol.append(lig.toMolecule())

        donors, acceptors = get_donors_acceptors(
            mol, exclude_water=False, exclude_backbone=False
        )

        wb = waterbridge_calculate(
            mol,
            donors,
            acceptors,
            "resname GOL",
            "protein and resname ASN and resid 155",
            order=1,
            dist_threshold=3.8,
            ignore_hs=True,
        )
        assert np.array_equal(wb, [[[3140, 2899, 2024]]]), wb

        wb = waterbridge_calculate(
            mol,
            donors,
            acceptors,
            "resname GOL",
            "protein and resname ASN and resid 155",
            order=2,
            dist_threshold=3.8,
            ignore_hs=True,
        )
        assert np.array_equal(wb, [[[3140, 2899, 2944, 2023], [3140, 2899, 2024]]]), wb

        wb = waterbridge_calculate(
            mol,
            donors,
            acceptors,
            "resname GOL",
            "protein",
            order=1,
            dist_threshold=3.8,
            ignore_hs=True,
        )
        assert np.array_equal(
            wb,
            [
                [
                    [3140, 2899, 2024],
                    [3142, 2857, 1317],
                    [3142, 2857, 2720],
                    [3142, 2857, 2737],
                    [3142, 2857, 2789],
                ]
            ],
        ), wb


if __name__ == "__main__":
    unittest.main(verbosity=2)


# cython: cdivision=True

import numpy as np
cimport numpy as np
from libcpp.vector cimport vector
from libc.math cimport round, sqrt, acos

# We now need to fix a datatype for our arrays. I've used the variable
# DTYPE for this, which is assigned to the usual NumPy runtime
# type info object.
UINT32 = np.uint32
FLOAT32 = np.float32

# "ctypedef" assigns a corresponding compile-time type to DTYPE_t. For
# every type in the numpy module there's a corresponding compile-time
# type with a _t-suffix.
ctypedef np.uint32_t UINT32_t
ctypedef np.float32_t FLOAT32_t

import cython
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef void _calculate_ring_mean(
        int f,
        int start_idx, 
        int end_idx, 
        np.ndarray[UINT32_t, ndim=1] rings_atoms, 
        np.ndarray[FLOAT32_t, ndim=3] coords,
        np.ndarray[FLOAT32_t, ndim=1] ring_mean,
    ):
    cdef FLOAT32_t coor
    for i in range(3):
        ring_mean[i] = 0

    for rr in range(start_idx, end_idx):
        for i in range(3):
            coor = coords[rings_atoms[rr], i, f]
            ring_mean[i] = ring_mean[i] + coor

    for i in range(3):
        ring_mean[i] = ring_mean[i] / (end_idx - start_idx)


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef float _wrapped_dist(
        int f,
        np.ndarray[FLOAT32_t, ndim=1] point1,
        np.ndarray[FLOAT32_t, ndim=1] point2,
        np.ndarray[FLOAT32_t, ndim=1] half_box,
        np.ndarray[FLOAT32_t, ndim=2] box,
    ):
    cdef FLOAT32_t val
    cdef float dist2 = 0
    for i in range(3):
        val = point1[i] - point2[i] 
        # Wrap the distance vector into the periodic box
        if abs(val) > half_box[i] and box[i, f] != 0:
            val = val - box[i, f] * round(val / box[i, f])
        dist2 = dist2 + (val * val)
    return dist2


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef void _cross_product(
        np.ndarray[FLOAT32_t, ndim=1] vec_a, 
        np.ndarray[FLOAT32_t, ndim=1] vec_b,
        np.ndarray[FLOAT32_t, ndim=1] res,
    ):
    res[0] = vec_a[1] * vec_b[2] - vec_a[2] * vec_b[1]
    res[1] = vec_a[2] * vec_b[0] - vec_a[0] * vec_b[2]
    res[2] = vec_a[0] * vec_b[1] - vec_a[1] * vec_b[0]

    cdef float vec_norm = 0 
    for i in range(3):
        vec_norm = vec_norm + (res[i] * res[i])
    vec_norm = sqrt(vec_norm)

    for i in range(3):
        res[i] = res[i] / vec_norm


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def calculate(
        np.ndarray[UINT32_t, ndim=1] rings_atoms, 
        np.ndarray[UINT32_t, ndim=1] rings1_start_indexes, 
        np.ndarray[UINT32_t, ndim=1] rings2_start_indexes, 
        np.ndarray[FLOAT32_t, ndim=3] coords, 
        np.ndarray[FLOAT32_t, ndim=2] box,
        float dist_threshold1=4.4,
        float angle_threshold1_max=30,
        float dist_threshold2=5.5,
        float angle_threshold2_min=60,
    ):
    cdef int r1, r2, f, i
    cdef int n_rings1 = rings1_start_indexes.shape[0] - 1  # Remove one for the index which we append to the start
    cdef int n_rings2 = rings2_start_indexes.shape[0] - 1
    cdef int n_frames = coords.shape[2]

    cdef vector[vector[int]] results
    cdef vector[vector[float]] distangles
    cdef np.ndarray[FLOAT32_t, ndim=1] ring1_mean = np.zeros(3, dtype=FLOAT32)
    cdef np.ndarray[FLOAT32_t, ndim=1] ring2_mean = np.zeros(3, dtype=FLOAT32)
    cdef np.ndarray[FLOAT32_t, ndim=1] ring1_normal = np.zeros(3, dtype=FLOAT32)
    cdef np.ndarray[FLOAT32_t, ndim=1] ring2_normal = np.zeros(3, dtype=FLOAT32)
    cdef np.ndarray[FLOAT32_t, ndim=1] half_box = np.zeros(3, dtype=FLOAT32)
    cdef np.ndarray[FLOAT32_t, ndim=1] tmp1 = np.zeros(3, dtype=FLOAT32)
    cdef np.ndarray[FLOAT32_t, ndim=1] tmp2 = np.zeros(3, dtype=FLOAT32)
    cdef FLOAT32_t dist2, dot_prod
    cdef int r1_start_idx, r2_start_idx, r1_end_idx, r2_end_idx

    dist_threshold1 = dist_threshold1 * dist_threshold1
    dist_threshold2 = dist_threshold2 * dist_threshold2

    for f in range(n_frames):
        results.push_back(vector[int]())
        distangles.push_back(vector[float]())
        for i in range(3):
            half_box[i] = box[i, f] / 2

        for r1 in range(n_rings1):
            for r2 in range(n_rings2):
                r1_start_idx = rings1_start_indexes[r1]
                r1_end_idx = rings1_start_indexes[r1+1]
                r2_start_idx = rings2_start_indexes[r2]
                r2_end_idx = rings2_start_indexes[r2+1]
                
                # Exit early if any of the two ring's atoms are too far
                for i in range(3):
                    tmp1[i] = coords[rings_atoms[r1_start_idx], i, f]
                    tmp2[i] = coords[rings_atoms[r2_start_idx], i, f]
                dist2 = _wrapped_dist(f, tmp1, tmp2, half_box, box)
                # Rings too far apart
                if dist2 > 225: # If two randomly picked atoms are more than 15A apart early-quit
                    continue

                # Calculate the ring centroids
                _calculate_ring_mean(f, r1_start_idx, r1_end_idx, rings_atoms, coords, ring1_mean)
                _calculate_ring_mean(f, r2_start_idx, r2_end_idx, rings_atoms, coords, ring2_mean)

                # Calculate the wrapped distance between the ring centroids
                dist2 = _wrapped_dist(f, ring1_mean, ring2_mean, half_box, box)

                # Ring centroids too far apart
                if dist2 > dist_threshold2:
                    # print(r1, r2, sqrt(dist2))
                    continue

                # Calculate the plane normals
                for i in range(3):
                    tmp1[i] = coords[rings_atoms[r1_start_idx], i, f] - coords[rings_atoms[r1_start_idx+2], i, f]
                    tmp2[i] = coords[rings_atoms[r1_start_idx+1], i, f] - coords[rings_atoms[r1_start_idx+2], i, f]
                _cross_product(tmp1, tmp2, ring1_normal)
                for i in range(3):
                    tmp1[i] = coords[rings_atoms[r2_start_idx], i, f] - coords[rings_atoms[r2_start_idx+2], i, f]
                    tmp2[i] = coords[rings_atoms[r2_start_idx+1], i, f] - coords[rings_atoms[r2_start_idx+2], i, f]
                _cross_product(tmp1, tmp2, ring2_normal)

                # Calculate angle between normals
                dot_prod = 0
                for i in range(3):
                    dot_prod = dot_prod + ring1_normal[i] * ring2_normal[i]
                angle = acos(dot_prod) * 57.29578  # Convert radians to degrees
                if angle > 90:  # minimal angle to line. necessary as perpendicular can point either way
                    angle = 180-angle

                # If dist < 4.4 and angle <= 30 deg
                # If dist < 5.5 and angle in range [60, 120]
                # print(r1, r2, sqrt(dist2), angle)
                if ((dist2 < dist_threshold1 and angle <= angle_threshold1_max) or 
                    (dist2 < dist_threshold2 and angle >= angle_threshold2_min)):
                    results[f].push_back(r1)
                    results[f].push_back(r2)
                    distangles[f].push_back(sqrt(dist2))
                    distangles[f].push_back(angle)

    return results, distangles





# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import moleculekit.vmdviewer
import io
from moleculekit.vmdviewer import getCurrentViewer


class VMDGraphicObject(object):
    """A superclass from which VMD graphic objects (e.g. the convex hull) inherit."""
    counter = 1

    def __init__(self, data):
        """Generic creation method. Not useful for the user."""
        self.n = self.counter
        self.data = data
        self.script = io.StringIO()
        self.valid = True
        VMDGraphicObject.counter += 1

    def delete(self):
        """Undisplay and delete a graphic object."""
        if not self.valid:
            raise Exception("The object has been deleted already.")

        n = self.n
        vmd = moleculekit.vmdviewer.getCurrentViewer()
        cmd = """
            set htmd_tmp $htmd_graphics_mol({0:d})
            foreach i $htmd_graphics({0:d}) {{ graphics $htmd_tmp delete $i }}
            unset htmd_graphics({0:d})
            unset htmd_graphics_mol({0:d})
        """.format(n)
        vmd.send(cmd)
        self.valid = False

    def _remember(self, s):
        # We can't get data back from VMD, so let it remember what's to be deleted
        n = self.n
        self.script.write("lappend htmd_graphics({:d}) [{:s}]\n".format(n, s))

    @staticmethod
    def tq(v):
        """Quote a numpy 3-vector to a TCL list."""
        return '{ ' + str(v).strip('[]') + ' }'


class VMDConvexHull(VMDGraphicObject):
    def __init__(self, mol, style="", preamble="", solid=False):
        """Display the convex hull of the given molecule.

        For preamble and color, see http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node129.html

        The function returns an instance of VMDGraphicsObject. To delete it, use the delete() method.

        Parameters
        ----------
        m: Molecule
            The object of which to show the hull (only 1 frame)
        style: str
            Style for wireframe lines
        preamble: str
            Commands (material, color) to be prefixed to the output.
            E.g.: "draw color red; graphics top materials on; graphics top material Transparent".
            Note that this affects later preamble-less commands.
        solid: bool
            Solid or wireframe

        Examples
        --------
        >>> from moleculekit.vmdgraphics import VMDConvexHull
        >>> m=Molecule("3PTB")
        >>> m.view()
        >>> mf=m.copy()
        >>> _ = mf.filter("protein ")
        >>> gh=VMDConvexHull(mf)  # doctest: +ELLIPSIS
        >>> gh.delete()
        """

        from scipy.spatial import ConvexHull

        if mol.coords.shape[2] != 1:
            raise Exception("Only one frame is supported")

        cc = mol.coords[:, :, 0]
        hull = ConvexHull(cc)

        super().__init__(hull)
        self.script.write(preamble + "\n")

        for i in range(hull.nsimplex):
            v1, v2, v3 = hull.simplices[i, :]
            c1s = VMDGraphicObject.tq(cc[v1, :])
            c2s = VMDGraphicObject.tq(cc[v2, :])
            c3s = VMDGraphicObject.tq(cc[v3, :])
            if solid:
                self._remember("draw triangle {:s} {:s} {:s}".format(c1s, c2s, c3s))
            else:
                self._remember("draw line {:s} {:s} {:s}".format(c1s, c2s, style))
                self._remember("draw line {:s} {:s} {:s}".format(c1s, c3s, style))
                self._remember("draw line {:s} {:s} {:s}".format(c2s, c3s, style))
                self.script.write("\n")

        self.script.write("set htmd_graphics_mol({:d}) [molinfo top]".format(self.n))
        cmd = self.script.getvalue()
        vmd = getCurrentViewer()
        vmd.send(cmd)


class VMDBox(VMDGraphicObject):
    def __init__(self, box, color='red'):
        """ Displays a box in VMD as lines of its edges.

        The function returns an instance of VMDGraphicsObject. To delete it, use the delete() method.

        Parameters
        ----------
        box : list
            The min and max positions of the edges. Given as [xmin, xmax, ymin, ymax, zmin, zmax]
        color : str
            Color of the lines
        """
        super().__init__(box)
        xmin, xmax, ymin, ymax, zmin, zmax = box
        mi = [xmin, ymin, zmin]
        ma = [xmax, ymax, zmax]

        self._remember('draw materials off')
        self._remember('draw color {}'.format(color))

        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(mi[0], mi[1], mi[2], ma[0], mi[1], mi[2]))
        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(mi[0], mi[1], mi[2], mi[0], ma[1], mi[2]))
        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(mi[0], mi[1], mi[2], mi[0], mi[1], ma[2]))

        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(ma[0], mi[1], mi[2], ma[0], ma[1], mi[2]))
        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(ma[0], mi[1], mi[2], ma[0], mi[1], ma[2]))

        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(mi[0], ma[1], mi[2], ma[0], ma[1], mi[2]))
        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(mi[0], ma[1], mi[2], mi[0], ma[1], ma[2]))

        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(mi[0], mi[1], ma[2], ma[0], mi[1], ma[2]))
        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(mi[0], mi[1], ma[2], mi[0], ma[1], ma[2]))

        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(ma[0], ma[1], ma[2], ma[0], ma[1], mi[2]))
        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(ma[0], ma[1], ma[2], mi[0], ma[1], ma[2]))
        self._remember('draw line "{} {} {}" "{} {} {}"\n'.format(ma[0], ma[1], ma[2], ma[0], mi[1], ma[2]))
        self.script.write("\n")
        self.script.write("set htmd_graphics_mol({:d}) [molinfo top]".format(self.n))
        cmd = self.script.getvalue()
        vmd = getCurrentViewer()
        vmd.send(cmd)


class VMDSphere(VMDGraphicObject):
    def __init__(self, xyz, color='red', radius=1):
        """ Displays a sphere in VMD.

        The function returns an instance of VMDGraphicsObject. To delete it, use the delete() method.

        Parameters
        ----------
        xyz : list
            The center of the sphere
        color : str
            Color of the sphere
        radius : float
            The radius of the sphere
        """
        super().__init__(xyz)
        #self._remember('draw materials off')
        self._remember('draw color {}'.format(color))

        self._remember('draw sphere "{} {} {}" radius {}\n'.format(xyz[0], xyz[1], xyz[2], radius))
        self.script.write("\n")
        self.script.write("set htmd_graphics_mol({:d}) [molinfo top]".format(self.n))
        cmd = self.script.getvalue()
        vmd = getCurrentViewer()
        vmd.send(cmd)


class VMDCylinder(VMDGraphicObject):
    def __init__(self, start, end, color='red', radius=1):
        """ Displays a cylinder in VMD.

        The function returns an instance of VMDGraphicsObject. To delete it, use the delete() method.

        Parameters
        ----------
        start : list
            The starting coordinates of the cylinder
        end : list
            The ending coordinates of the cylinder
        color : str
            Color of the cylinder
        radius : float
            The radius of the cylinder
        """
        super().__init__([start, end])
        #self._remember('draw materials off')
        self._remember('draw color {}'.format(color))
        self._remember('draw cylinder {{ {} }} {{ {} }} radius {}\n'.format(' '.join(map(str, start)), ' '.join(map(str, end)), radius))
        self.script.write("\n")
        self.script.write("set htmd_graphics_mol({:d}) [molinfo top]".format(self.n))
        cmd = self.script.getvalue()
        vmd = getCurrentViewer()
        vmd.send(cmd)


class VMDText(VMDGraphicObject):
    def __init__(self, text, xyz, color='red'):
        """ Displays a text in VMD.

        The function returns an instance of VMDGraphicsObject. To delete it, use the delete() method.

        Parameters
        ----------
        text : str
            The text
        xyz : list
            The position of the text
        color : str
            Color of the text
        """
        super().__init__(xyz)
        self._remember('draw materials off')
        self._remember('draw color {}'.format(color))

        self._remember('draw text "{} {} {}" "{}"\n'.format(xyz[0], xyz[1], xyz[2], text))
        self.script.write("\n")
        self.script.write("set htmd_graphics_mol({:d}) [molinfo top]".format(self.n))
        cmd = self.script.getvalue()
        vmd = getCurrentViewer()
        vmd.send(cmd)


class VMDIsosurface(VMDGraphicObject):
    def __init__(self, arr, vecMin, vecRes, color=8, isovalue=0.5, name=None, draw='solid'):
        """ Displays an isosurface in VMD

        The function returns an instance of VMDGraphicsObject. To delete it, use the delete() method.

        Parameters
        ----------
        arr: np.ndarray
            3D array with volumetric data.
        vecMin: np.ndarray
            3D vector denoting the minimal corner of the grid
        vecRes: np.ndarray
            3D vector denoting the resolution of the grid in each dimension
        color: str
            The color to be used for the isosurface
        name: str
            A name for the representation
        draw: str ('solid', 'wireframe')
            Drawing style for the isosurface
        """
        super().__init__(arr)
        from moleculekit.util import tempname, writeVoxels
        import numpy as np
        import os

        filename = tempname(suffix='.cube')
        writeVoxels(arr, filename, vecMin, vecRes)

        vmd = getCurrentViewer()

        drawmapping = {'solid': 0, 'wireframe': 1}

        vmd.send('mol new {} type cube first 0 last -1 step 1 waitfor 1 volsets {{0 }}'.format(filename))
        vmd.send('mol modstyle top top Isosurface {} 0 2 {} 2 1'.format(isovalue, drawmapping[draw]))
        vmd.send('mol modcolor top top ColorID {}'.format(color))
        vmd.send('set htmd_graphics_mol({:d}) [molinfo top]'.format(self.n))
        if name is not None:
            vmd.send('mol rename top {{{}}}'.format(name))

        if os.path.exists(filename):
            os.unlink(filename)

    def delete(self):
        vmd = getCurrentViewer()
        vmd.send('mol delete htmd_graphics_mol({:d})'.format(self.n))


class VMDLabels(VMDGraphicObject):
    count = 0
    def __init__(self, mol, selection, molid='top', textsize=0.5, textcolor='green'):
        """ Displays labels on atoms in VMD.

        Examples
        --------
        >>> from moleculekit.molecule import Molecule  # doctest: +ELLIPSIS
        >>> from moleculekit.vmdgraphics import VMDLabels
        >>> mol = Molecule('3ptb')
        >>> mol.view()
        >>> x = VMDLabels(mol, 'resid 40')
        >>> y = VMDLabels(mol, 'resid 50')
        >>> y.delete()
        >>> x.delete()
        """
        # TODO: After deleting labels it breaks. Better don't delete or fix the code
        super().__init__(None)
        idx = mol.atomselect(selection, indexes=True)
        cmd = '''label textsize {s}
color Labels Atoms {c}
set molnum [molinfo {molid}]
set i {start}
foreach n [list {idx}] {{
    label add Atoms $molnum/$n
    label textformat Atoms $i {{%a}}
    incr i
}}'''.format(s=textsize, c=textcolor, sel=selection, molid=molid, idx=' '.join(map(str, idx)), start=VMDLabels.count)
        self.labelidx = list(range(VMDLabels.count, VMDLabels.count+len(idx)))
        VMDLabels.count += len(idx)
        vmd = getCurrentViewer()
        vmd.send(cmd)
        # print(cmd)

    def delete(self):
        vmd = getCurrentViewer()
        for i in self.labelidx[::-1]:
            vmd.send('label delete Atoms {}'.format(i))
            VMDLabels.count -= 1


import unittest

# Practically only testing execution of the classes
class _TestVMDGraphics(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.molecule import Molecule

        self.viewer = getCurrentViewer(dispdev='text')
        self.mol3PTB = Molecule('3PTB')

    def test_convex_hull(self):
        VMDConvexHull(self.mol3PTB)
        VMDConvexHull(self.mol3PTB, solid=True)

    def test_box(self):
        VMDBox([0, 10, -5, 3, 100, 400])

    def test_text(self):
        VMDText('test text', [10, -3.4, 6])

    def test_isosurface(self):
        import numpy as np
        VMDIsosurface(np.random.rand(3, 4, 2), [-1, 25, 3], [1, 1, 1])

    def test_labels(self):
        VMDLabels(self.mol3PTB, 'resname BEN')

    def test_sphere(self):
        VMDSphere([0, -5.3, 195], radius=6)


if __name__ == "__main__":
    """
    from moleculekit.molecule import Molecule
    import moleculekit.vmdgraphics

    m=Molecule("3PTB")
    m.view()
    mf=m.copy()
    mf.filter("protein")
    mgh=moleculekit.vmdgraphics.VMDConvexHull(mf)

    n=Molecule("1JNO")
    n.view()
    nf=n.copy()
    nf.filter("protein")
    ngh=moleculekit.vmdgraphics.VMDConvexHull(nf,solid=True)

    """
    unittest.main(verbosity=2)
    # from moleculekit.molecule import Molecule
    # import doctest
    # doctest.testmod()
